# Copyright 2017 Redhat.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import ipaddress
import os
import shutil
from unittest import mock

from oslo_config import fixture as oslo_fixture

from octavia.amphorae.backends.agent.api_server import osutils
from octavia.common import config
from octavia.common import constants as consts
from octavia.common import exceptions as octavia_exceptions
from octavia.common import utils
from octavia.tests.common import utils as test_utils
from octavia.tests.unit import base


class TestOSUtils(base.TestCase):

    def setUp(self):
        super().setUp()

        self.base_os_util = osutils.BaseOS('unknown')

        with mock.patch('distro.id',
                        return_value='ubuntu'):
            self.ubuntu_os_util = osutils.BaseOS.get_os_util()

        with mock.patch('distro.id',
                        return_value='rhel'):
            self.rh_os_util = osutils.BaseOS.get_os_util()

        with mock.patch('distro.id', return_value='centos'):
            with mock.patch('distro.version', return_value='8'):
                self.centos_os_util = osutils.BaseOS.get_os_util()

        with mock.patch('distro.id', return_value='centos'):
            with mock.patch('distro.version', return_value='7'):
                self.centos7_os_util = osutils.BaseOS.get_os_util()

    def test_get_os_util(self):
        with mock.patch('distro.id',
                        return_value='ubuntu'):
            returned_cls = osutils.BaseOS.get_os_util()
            self.assertIsInstance(returned_cls, osutils.Ubuntu)
        with mock.patch('distro.id',
                        return_value='fedora'):
            returned_cls = osutils.BaseOS.get_os_util()
            self.assertIsInstance(returned_cls, osutils.RH)
        with mock.patch('distro.id',
                        return_value='rhel'):
            returned_cls = osutils.BaseOS.get_os_util()
            self.assertIsInstance(returned_cls, osutils.RH)
        with mock.patch('distro.id',
                        return_value='centos'):
            returned_cls = osutils.BaseOS.get_os_util()
            self.assertIsInstance(returned_cls, osutils.CentOS)
        with mock.patch('distro.id',
                        return_value='FakeOS'):
            self.assertRaises(
                octavia_exceptions.InvalidAmphoraOperatingSystem,
                osutils.BaseOS.get_os_util)

    def test_get_network_interface_file(self):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))

        fake_agent_server_network_dir = "/path/to/interface"
        fake_agent_server_network_file = "/path/to/interfaces_file"

        base_fake_nic_path = os.path.join(fake_agent_server_network_dir,
                                          consts.NETNS_PRIMARY_INTERFACE)
        base_real_nic_path = os.path.join(
            consts.UBUNTU_AMP_NET_DIR_TEMPLATE.format(
                netns=consts.AMPHORA_NAMESPACE),
            consts.NETNS_PRIMARY_INTERFACE)

        rh_interface_name = 'ifcfg-{nic}'.format(
            nic=consts.NETNS_PRIMARY_INTERFACE)
        rh_fake_nic_path = os.path.join(fake_agent_server_network_dir,
                                        rh_interface_name)
        rh_real_nic_path = os.path.join(
            consts.RH_AMP_NET_DIR_TEMPLATE.format(
                netns=consts.AMPHORA_NAMESPACE),
            rh_interface_name)

        ubuntu_interface_name = '{nic}.cfg'.format(
            nic=consts.NETNS_PRIMARY_INTERFACE)
        ubuntu_fake_nic_path = os.path.join(fake_agent_server_network_dir,
                                            ubuntu_interface_name)
        ubuntu_real_nic_path = os.path.join(
            consts.UBUNTU_AMP_NET_DIR_TEMPLATE.format(
                netns=consts.AMPHORA_NAMESPACE),
            ubuntu_interface_name)

        # Check that agent_server_network_file is returned, when provided
        conf.config(group="amphora_agent",
                    agent_server_network_file=fake_agent_server_network_file)

        base_interface_file = (
            self.base_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(fake_agent_server_network_file, base_interface_file)

        rh_interface_file = (
            self.rh_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(fake_agent_server_network_file, rh_interface_file)

        ubuntu_interface_file = (
            self.ubuntu_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(fake_agent_server_network_file, ubuntu_interface_file)

        # Check that agent_server_network_dir is used, when provided
        conf.config(group="amphora_agent", agent_server_network_file=None)
        conf.config(group="amphora_agent",
                    agent_server_network_dir=fake_agent_server_network_dir)

        base_interface_file = (
            self.base_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(base_fake_nic_path, base_interface_file)

        rh_interface_file = (
            self.rh_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(rh_fake_nic_path, rh_interface_file)

        ubuntu_interface_file = (
            self.ubuntu_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(ubuntu_fake_nic_path, ubuntu_interface_file)

        # Check When neither agent_server_network_dir or
        # agent_server_network_file where provided.
        conf.config(group="amphora_agent", agent_server_network_file=None)
        conf.config(group="amphora_agent", agent_server_network_dir=None)

        base_interface_file = (
            self.base_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(base_real_nic_path, base_interface_file)

        rh_interface_file = (
            self.rh_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(rh_real_nic_path, rh_interface_file)

        ubuntu_interface_file = (
            self.ubuntu_os_util.
            get_network_interface_file(consts.NETNS_PRIMARY_INTERFACE))
        self.assertEqual(ubuntu_real_nic_path, ubuntu_interface_file)

    def _test_RH_get_static_routes_interface_file(self, version):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))

        fake_agent_server_network_dir = "/path/to/interface"
        fake_agent_server_network_file = "/path/to/interfaces_file"

        route = 'route6' if version == 6 else 'route'
        rh_route_name = '{route}-{nic}'.format(
            route=route, nic=consts.NETNS_PRIMARY_INTERFACE)
        rh_fake_route_path = os.path.join(fake_agent_server_network_dir,
                                          rh_route_name)
        rh_real_route_path = os.path.join(
            consts.RH_AMP_NET_DIR_TEMPLATE.format(
                netns=consts.AMPHORA_NAMESPACE),
            rh_route_name)

        # Check that agent_server_network_file is returned, when provided
        conf.config(group="amphora_agent",
                    agent_server_network_file=fake_agent_server_network_file)

        rh_route_file = (
            self.rh_os_util.
            get_static_routes_interface_file(consts.NETNS_PRIMARY_INTERFACE,
                                             version))
        self.assertEqual(fake_agent_server_network_file, rh_route_file)

        # Check that agent_server_network_dir is used, when provided
        conf.config(group="amphora_agent", agent_server_network_file=None)
        conf.config(group="amphora_agent",
                    agent_server_network_dir=fake_agent_server_network_dir)

        rh_route_file = (
            self.rh_os_util.
            get_static_routes_interface_file(consts.NETNS_PRIMARY_INTERFACE,
                                             version))
        self.assertEqual(rh_fake_route_path, rh_route_file)

        # Check When neither agent_server_network_dir or
        # agent_server_network_file where provided.
        conf.config(group="amphora_agent", agent_server_network_file=None)
        conf.config(group="amphora_agent", agent_server_network_dir=None)

        rh_route_file = (
            self.rh_os_util.
            get_static_routes_interface_file(consts.NETNS_PRIMARY_INTERFACE,
                                             version))
        self.assertEqual(rh_real_route_path, rh_route_file)

    def test_RH_get_static_routes_interface_file(self):
        self._test_RH_get_static_routes_interface_file(4)

    def test_RH_get_static_routes_interface_file_ipv6(self):
        self._test_RH_get_static_routes_interface_file(6)

    def _test_RH_get_route_rules_interface_file(self, version):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))

        fake_agent_server_network_dir = "/path/to/interface"
        fake_agent_server_network_file = "/path/to/interfaces_file"

        rule = 'rule6' if version == 6 else 'rule'
        rh_route_rules_name = '{rule}-{nic}'.format(
            rule=rule, nic=consts.NETNS_PRIMARY_INTERFACE)
        rh_fake_route_rules_path = os.path.join(fake_agent_server_network_dir,
                                                rh_route_rules_name)
        rh_real_route_rules_path = os.path.join(
            consts.RH_AMP_NET_DIR_TEMPLATE.format(
                netns=consts.AMPHORA_NAMESPACE),
            rh_route_rules_name)

        # Check that agent_server_network_file is returned, when provided
        conf.config(group="amphora_agent",
                    agent_server_network_file=fake_agent_server_network_file)

        rh_route_rules_file = (
            self.rh_os_util.
            get_route_rules_interface_file(consts.NETNS_PRIMARY_INTERFACE,
                                           version))
        self.assertEqual(fake_agent_server_network_file, rh_route_rules_file)

        # Check that agent_server_network_dir is used, when provided
        conf.config(group="amphora_agent", agent_server_network_file=None)
        conf.config(group="amphora_agent",
                    agent_server_network_dir=fake_agent_server_network_dir)

        rh_route_rules_file = (
            self.rh_os_util.
            get_route_rules_interface_file(consts.NETNS_PRIMARY_INTERFACE,
                                           version))
        self.assertEqual(rh_fake_route_rules_path, rh_route_rules_file)

        # Check When neither agent_server_network_dir or
        # agent_server_network_file where provided.
        conf.config(group="amphora_agent", agent_server_network_file=None)
        conf.config(group="amphora_agent", agent_server_network_dir=None)

        rh_route_rules_file = (
            self.rh_os_util.
            get_route_rules_interface_file(consts.NETNS_PRIMARY_INTERFACE,
                                           version))
        self.assertEqual(rh_real_route_rules_path, rh_route_rules_file)

    def test_RH_get_route_rules_interface_file(self):
        self._test_RH_get_route_rules_interface_file(4)

    def test_RH_get_route_rules_interface_file_ipv6(self):
        self._test_RH_get_route_rules_interface_file(6)

    def test_cmd_get_version_of_installed_package(self):
        package_name = 'foo'
        ubuntu_cmd = "dpkg-query -W -f=${{Version}} {name}".format(
            name=package_name)
        rh_cmd = "rpm -q --queryformat %{{VERSION}} {name}".format(
            name=package_name)

        returned_ubuntu_cmd = (
            self.ubuntu_os_util.cmd_get_version_of_installed_package(
                package_name))
        self.assertEqual(ubuntu_cmd, returned_ubuntu_cmd)

        returned_rh_cmd = (self.rh_os_util.
                           cmd_get_version_of_installed_package(package_name))
        self.assertEqual(rh_cmd, returned_rh_cmd)

    def test_cmd_get_version_of_installed_package_mapped(self):
        package_name = 'haproxy'
        centos7_cmd = "rpm -q --queryformat %{VERSION} haproxy18"

        returned_centos7_cmd = (
            self.centos7_os_util.cmd_get_version_of_installed_package(
                package_name))
        self.assertEqual(centos7_cmd, returned_centos7_cmd)

        centos_cmd = "rpm -q --queryformat %{VERSION} haproxy"
        returned_centos_cmd = (
            self.centos_os_util.cmd_get_version_of_installed_package(
                package_name))
        self.assertEqual(centos_cmd, returned_centos_cmd)

    def test_has_ifup_all(self):
        self.assertTrue(self.base_os_util.has_ifup_all())
        self.assertTrue(self.ubuntu_os_util.has_ifup_all())
        self.assertFalse(self.rh_os_util.has_ifup_all())

    def test_write_vip_interface_file(self):
        netns_interface = u'eth1234'
        FIXED_IP = u'192.0.2.2'
        SUBNET_CIDR = u'192.0.2.0/24'
        GATEWAY = u'192.51.100.1'
        DEST1 = u'198.51.100.0/24'
        DEST2 = u'203.0.113.0/24'
        NEXTHOP = u'192.0.2.1'
        MTU = 1450
        FIXED_IP_IPV6 = u'2001:0db8:0000:0000:0000:0000:0000:0001'
        # Subnet prefix is purposefully not 32, because that coincidentally
        # matches the result of any arbitrary IPv4->prefixlen conversion
        SUBNET_CIDR_IPV6 = u'2001:db8::/70'

        ip = ipaddress.ip_address(FIXED_IP)
        network = ipaddress.ip_network(SUBNET_CIDR)
        broadcast = network.broadcast_address.exploded
        netmask = network.netmask.exploded
        netmask_prefix = utils.netmask_to_prefix(netmask)

        ipv6 = ipaddress.ip_address(FIXED_IP_IPV6)
        networkv6 = ipaddress.ip_network(SUBNET_CIDR_IPV6)
        broadcastv6 = networkv6.broadcast_address.exploded
        netmaskv6 = networkv6.prefixlen

        host_routes = [
            {'gw': NEXTHOP, 'network': ipaddress.ip_network(DEST1)},
            {'gw': NEXTHOP, 'network': ipaddress.ip_network(DEST2)}
        ]

        path = self.ubuntu_os_util.get_network_interface_file(netns_interface)
        mock_open = self.useFixture(test_utils.OpenFixture(path)).mock_open
        mock_template = mock.MagicMock()

        # Test an IPv4 VIP
        with mock.patch('os.open'), mock.patch.object(
                os, 'fdopen', mock_open):
            self.ubuntu_os_util.write_vip_interface_file(
                interface_file_path=path,
                primary_interface=netns_interface,
                vip=FIXED_IP,
                ip=ip,
                broadcast=broadcast,
                netmask=netmask,
                gateway=GATEWAY,
                mtu=MTU,
                vrrp_ip=None,
                vrrp_version=None,
                render_host_routes=host_routes,
                template_vip=mock_template)

        mock_template.render.assert_called_once_with(
            consts=consts,
            interface=netns_interface,
            vip=FIXED_IP,
            vip_ipv6=False,
            prefix=netmask_prefix,
            broadcast=broadcast,
            netmask=netmask,
            gateway=GATEWAY,
            network=SUBNET_CIDR,
            mtu=MTU,
            vrrp_ip=None,
            vrrp_ipv6=False,
            host_routes=host_routes,
            topology="SINGLE",
        )

        # Now test with an IPv6 VIP
        mock_template.reset_mock()
        with mock.patch('os.open'), mock.patch.object(
                os, 'fdopen', mock_open):
            self.ubuntu_os_util.write_vip_interface_file(
                interface_file_path=path,
                primary_interface=netns_interface,
                vip=FIXED_IP_IPV6,
                ip=ipv6,
                broadcast=broadcastv6,
                netmask=netmaskv6,
                gateway=GATEWAY,
                mtu=MTU,
                vrrp_ip=None,
                vrrp_version=None,
                render_host_routes=host_routes,
                template_vip=mock_template)

        mock_template.render.assert_called_once_with(
            consts=consts,
            interface=netns_interface,
            vip=FIXED_IP_IPV6,
            vip_ipv6=True,
            prefix=netmaskv6,
            broadcast=broadcastv6,
            netmask=netmaskv6,
            gateway=GATEWAY,
            network=SUBNET_CIDR_IPV6,
            mtu=MTU,
            vrrp_ip=None,
            vrrp_ipv6=False,
            host_routes=host_routes,
            topology="SINGLE",
        )

    def test_write_port_interface_file(self):
        FIXED_IP = u'192.0.2.2'
        NEXTHOP = u'192.0.2.1'
        DEST = u'198.51.100.0/24'
        host_routes = [
            {'nexthop': NEXTHOP, 'destination': ipaddress.ip_network(DEST)}
        ]
        FIXED_IP_IPV6 = u'2001:db8::2'
        NEXTHOP_IPV6 = u'2001:db8::1'
        DEST_IPV6 = u'2001:db8:51:100::/64'
        host_routes_ipv6 = [
            {'nexthop': NEXTHOP_IPV6,
             'destination': ipaddress.ip_network(DEST_IPV6)}
        ]
        ip_addr = {'ip_address': FIXED_IP, 'host_routes': host_routes}
        ipv6_addr = {'ip_address': FIXED_IP_IPV6,
                     'host_routes': host_routes_ipv6}

        netns_interface = 'eth1234'
        MTU = 1450
        fixed_ips = [ip_addr, ipv6_addr]
        path = 'mypath'
        mock_template = mock.MagicMock()
        mock_open = self.useFixture(test_utils.OpenFixture(path)).mock_open
        mock_gen_text = mock.MagicMock()
        mock_local_scripts = mock.MagicMock()
        mock_wr_fi = mock.MagicMock()

        with mock.patch('os.open'), mock.patch.object(
                os, 'fdopen', mock_open), mock.patch.object(
                osutils.BaseOS, '_generate_network_file_text', mock_gen_text):
            self.base_os_util.write_port_interface_file(
                netns_interface=netns_interface,
                fixed_ips=fixed_ips,
                mtu=MTU,
                interface_file_path=path,
                template_port=mock_template)

        mock_gen_text.assert_called_once_with(
            netns_interface, fixed_ips, MTU, mock_template)

        mock_gen_text.reset_mock()

        with mock.patch('os.open'), mock.patch.object(
                os, 'fdopen', mock_open), mock.patch.object(
                osutils.BaseOS, '_generate_network_file_text',
                mock_gen_text), mock.patch.object(
                osutils.RH, '_write_ifup_ifdown_local_scripts_if_possible',
                mock_local_scripts), mock.patch.object(
                osutils.RH, 'write_static_routes_interface_file', mock_wr_fi):
            self.rh_os_util.write_port_interface_file(
                netns_interface=netns_interface,
                fixed_ips=fixed_ips,
                mtu=MTU,
                interface_file_path=path,
                template_port=mock_template)

        rh_route_name = 'route-{nic}'.format(nic=netns_interface)
        rh_real_route_path = os.path.join(
            consts.RH_AMP_NET_DIR_TEMPLATE.format(
                netns=consts.AMPHORA_NAMESPACE),
            rh_route_name)
        rh_route_name_ipv6 = 'route6-{nic}'.format(nic=netns_interface)
        rh_real_route_path_ipv6 = os.path.join(
            consts.RH_AMP_NET_DIR_TEMPLATE.format(
                netns=consts.AMPHORA_NAMESPACE),
            rh_route_name_ipv6)

        exp_routes = [
            {'network': ipaddress.ip_network(DEST), 'gw': NEXTHOP}
        ]
        exp_routes_ipv6 = [
            {'network': ipaddress.ip_network(DEST_IPV6), 'gw': NEXTHOP_IPV6}
        ]
        expected_calls = [
            mock.call(rh_real_route_path, netns_interface,
                      exp_routes, mock.ANY, None, None, None),
            mock.call(rh_real_route_path_ipv6, netns_interface,
                      exp_routes_ipv6, mock.ANY, None, None, None)]

        mock_gen_text.assert_called_once_with(
            netns_interface, fixed_ips, MTU, mock_template)
        self.assertEqual(2, mock_wr_fi.call_count)
        mock_wr_fi.assert_has_calls(expected_calls)
        mock_local_scripts.assert_called_once()

    @mock.patch('shutil.copy2')
    @mock.patch('os.makedirs')
    @mock.patch('shutil.copytree')
    def test_create_netns_dir(self, mock_copytree, mock_makedirs, mock_copy2):
        network_dir = 'foo'
        netns_network_dir = 'fake_netns_network'
        ignore = shutil.ignore_patterns('fake_eth*', 'fake_loopback*')
        self.rh_os_util.create_netns_dir(network_dir,
                                         netns_network_dir,
                                         ignore)
        mock_copytree.assert_any_call(
            network_dir,
            os.path.join('/etc/netns/',
                         consts.AMPHORA_NAMESPACE,
                         netns_network_dir),
            ignore=ignore,
            symlinks=True)

        mock_makedirs.assert_any_call(os.path.join('/etc/netns/',
                                                   consts.AMPHORA_NAMESPACE))
        mock_copy2.assert_any_call(
            '/etc/sysconfig/network',
            '/etc/netns/{netns}/sysconfig'.format(
                netns=consts.AMPHORA_NAMESPACE))

        mock_copytree.reset_mock()
        mock_makedirs.reset_mock()
        mock_copy2.reset_mock()

        self.ubuntu_os_util.create_netns_dir(network_dir,
                                             netns_network_dir,
                                             ignore)
        mock_copytree.assert_any_call(
            network_dir,
            os.path.join('/etc/netns/',
                         consts.AMPHORA_NAMESPACE,
                         netns_network_dir),
            ignore=ignore,
            symlinks=True)

        mock_makedirs.assert_any_call(os.path.join('/etc/netns/',
                                                   consts.AMPHORA_NAMESPACE))
        mock_copy2.assert_not_called()



#    Copyright (c) 2019 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add protocol in listener keys

Revision ID: a5762a99609a
Revises: 392fb85b4419
Create Date: 2019-06-28 14:02:11.415292

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = 'a5762a99609a'
down_revision = '392fb85b4419'


def upgrade():
    op.execute("ALTER TABLE `listener` "
               "DROP INDEX `uq_listener_load_balancer_id_protocol_port`, "
               "ADD UNIQUE KEY "
               "`uq_listener_load_balancer_id_protocol_port` "
               "(`load_balancer_id`, `protocol`, `protocol_port`)")



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from wsme import exc
from wsme.rest import json as wsme_json
from wsme import types as wsme_types

from octavia.api.v2.types import health_monitor as hm_type
from octavia.common import constants
from octavia.tests.unit.api.v2.types import base


class TestHealthMonitor(object):

    _type = None

    def test_invalid_type(self):
        body = {"delay": 1, "timeout": 1, "max_retries": 1}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": 1, "pool_id": uuidutils.generate_uuid()})
            self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                              body)

    def test_invalid_delay(self):
        body = {"delay": "one", "timeout": 1, "max_retries": 1}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(ValueError, wsme_json.fromjson, self._type, body)

    def test_invalid_timeout(self):
        body = {"delay": 1, "timeout": "one", "max_retries": 1}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(ValueError, wsme_json.fromjson, self._type, body)

    def test_invalid_max_retries_down(self):
        body = {"delay": 1, "timeout": 1, "max_retries": "one"}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(ValueError, wsme_json.fromjson, self._type, body)

    def test_invalid_max_retries(self):
        body = {"delay": 1, "timeout": 1, "max_retries": "one"}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(ValueError, wsme_json.fromjson, self._type, body)

    def test_invalid_http_method(self):
        body = {"delay": 1, "timeout": 1, "max_retries": 1,
                "http_method": 1}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                          body)

    def test_invalid_url_path(self):
        body = {"delay": 1, "timeout": 1, "max_retries": 1, "url_path": 1}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson,
                          self._type, body)

    def test_invalid_url_path_with_url(self):
        body = {"delay": 1, "timeout": 1, "max_retries": 1,
                "url_path": 'https://www.openstack.org'}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson,
                          self._type, body)

    def test_invalid_url_path_no_leading_slash(self):
        body = {"delay": 1, "timeout": 1, "max_retries": 1,
                "url_path": 'blah'}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson,
                          self._type, body)

    def test_invalid_expected_codes(self):
        body = {"delay": 1, "timeout": 1, "max_retries": 1,
                "expected_codes": "lol"}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid()})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                          body)

    def test_invalid_tags(self):
        body = {"tags": "invalid_tag"}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid(),
                         "delay": 1, "timeout": 1, "max_retries": 1})
        self.assertRaises(ValueError, wsme_json.fromjson, self._type,
                          body)
        body = {"tags": [1, 2]}
        if self._type is hm_type.HealthMonitorPOST:
            body.update({"type": constants.PROTOCOL_HTTP,
                         "pool_id": uuidutils.generate_uuid(),
                         "delay": 1, "timeout": 1, "max_retries": 1})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                          body)


class TestHealthMonitorPOST(base.BaseTypesTest, TestHealthMonitor):

    _type = hm_type.HealthMonitorPOST

    def test_health_monitor(self):
        body = {"type": constants.HEALTH_MONITOR_HTTP, "delay": 1,
                "timeout": 1, "max_retries_down": 1, "max_retries": 1,
                "pool_id": uuidutils.generate_uuid(),
                "tags": ['test_tag']}
        hm = wsme_json.fromjson(self._type, body)
        self.assertTrue(hm.admin_state_up)

    def test_type_mandatory(self):
        body = {"delay": 80, "timeout": 1, "max_retries": 1,
                "pool_id": uuidutils.generate_uuid()}
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                          body)

    def test_delay_mandatory(self):
        body = {"type": constants.HEALTH_MONITOR_HTTP, "timeout": 1,
                "max_retries": 1, "pool_id": uuidutils.generate_uuid()}
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                          body)

    def test_timeout_mandatory(self):
        body = {"type": constants.HEALTH_MONITOR_HTTP, "delay": 1,
                "max_retries": 1, "pool_id": uuidutils.generate_uuid()}
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                          body)

    def test_max_retries_mandatory(self):
        body = {"type": constants.HEALTH_MONITOR_HTTP, "delay": 1,
                "timeout": 1, "pool_id": uuidutils.generate_uuid()}
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson, self._type,
                          body)

    def test_default_health_monitor_values(self):
        # http_method = 'GET'
        # url_path = '/'
        # expected_codes = '200'
        # max_retries_down = 3
        # admin_state_up = True
        # The above are not required but should have the above example defaults
        body = {"type": constants.HEALTH_MONITOR_HTTP, "delay": 1,
                "timeout": 1, "max_retries": 1,
                "pool_id": uuidutils.generate_uuid()}
        hmpost = wsme_json.fromjson(self._type, body)
        self.assertEqual(wsme_types.Unset, hmpost.http_method)
        self.assertEqual(wsme_types.Unset, hmpost.url_path)
        self.assertEqual(wsme_types.Unset, hmpost.expected_codes)
        self.assertEqual(3, hmpost.max_retries_down)
        self.assertTrue(hmpost.admin_state_up)

    def test_url_path_with_query_and_fragment(self):
        url_path = "/v2/index?a=12,b=34#123dd"
        body = {"type": constants.HEALTH_MONITOR_HTTP, "delay": 1,
                "timeout": 1, "max_retries": 1,
                "pool_id": uuidutils.generate_uuid(),
                "url_path": url_path}
        hmpost = wsme_json.fromjson(self._type, body)
        self.assertEqual(wsme_types.Unset, hmpost.http_method)
        self.assertEqual(url_path, hmpost.url_path)
        self.assertEqual(wsme_types.Unset, hmpost.expected_codes)
        self.assertEqual(3, hmpost.max_retries_down)
        self.assertTrue(hmpost.admin_state_up)

    def test_non_uuid_project_id(self):
        body = {"type": constants.HEALTH_MONITOR_HTTP, "delay": 1,
                "timeout": 1, "max_retries_down": 1, "max_retries": 1,
                "project_id": "non-uuid",
                "pool_id": uuidutils.generate_uuid()}
        hm = wsme_json.fromjson(self._type, body)
        self.assertEqual(hm.project_id, body['project_id'])


class TestHealthMonitorPUT(base.BaseTypesTest, TestHealthMonitor):

    _type = hm_type.HealthMonitorPUT

    def test_health_monitor(self):
        body = {"http_method": constants.HEALTH_MONITOR_HTTP_METHOD_HEAD,
                "tags": ['test_tag']}
        hm = wsme_json.fromjson(self._type, body)
        self.assertEqual(wsme_types.Unset, hm.admin_state_up)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from octavia.db import api as db_api
from octavia.db import repositories as repo
from octavia.statistics import stats_base

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class StatsUpdateDb(stats_base.StatsDriverMixin):

    def __init__(self):
        super().__init__()
        self.listener_stats_repo = repo.ListenerStatisticsRepository()

    def update_stats(self, listener_stats, deltas=False):
        """This function is to update the db with listener stats"""
        session = db_api.get_session()
        for stats_object in listener_stats:
            LOG.debug("Updating listener stats in db for listener `%s` / "
                      "amphora `%s`: %s",
                      stats_object.listener_id, stats_object.amphora_id,
                      stats_object.get_stats())
            if deltas:
                self.listener_stats_repo.increment(session, stats_object)
            else:
                self.listener_stats_repo.replace(session, stats_object)



#  Copyright 2019 Red Hat, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_log import log as logging
from taskflow import retry

LOG = logging.getLogger(__name__)


class SleepingRetryTimesController(retry.Times):
    """A retry controller to attempt subflow retries a number of times.

    This retry controller overrides the Times on_failure to inject a
    sleep interval between retries.
    It also adds a log message when all of the retries are exhausted.

    :param attempts: number of attempts to retry the associated subflow
                        before giving up
    :type attempts: int
    :param name: Meaningful name for this atom, should be something that is
                 distinguishable and understandable for notification,
                 debugging, storing and any other similar purposes.
    :param provides: A set, string or list of items that
                     this will be providing (or could provide) to others, used
                     to correlate and associate the thing/s this atom
                     produces, if it produces anything at all.
    :param requires: A set or list of required inputs for this atom's
                     ``execute`` method.
    :param rebind: A dict of key/value pairs used to define argument
                   name conversions for inputs to this atom's ``execute``
                   method.
    :param revert_all: when provided this will cause the full flow to revert
                       when the number of attempts that have been tried
                       has been reached (when false, it will only locally
                       revert the associated subflow)
    :type revert_all: bool
    :param interval: Interval, in seconds, between retry attempts.
    :type interval: int
    """

    def __init__(self, attempts=1, name=None, provides=None, requires=None,
                 auto_extract=True, rebind=None, revert_all=False, interval=1):
        super().__init__(attempts, name, provides, requires, auto_extract,
                         rebind, revert_all)
        self._interval = interval

    def on_failure(self, history, *args, **kwargs):
        if len(history) < self._attempts:
            LOG.warning('%s attempt %s of %s failed. Sleeping %s seconds and '
                        'retrying.',
                        self.name[self.name.startswith('retry-') and
                                  len('retry-'):], len(history),
                        self._attempts, self._interval)
            time.sleep(self._interval)
            return retry.RETRY
        return self._revert_action

    def revert(self, history, *args, **kwargs):
        LOG.error('%s retries with interval %s seconds have failed for %s. '
                  'Giving up.', len(history), self._interval, self.name)



# Copyright (c) 2018 China Telecom Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add l7policy and l7rule quota

Revision ID: 32e5c35b26a8
Revises: d3c8a090f3de
Create Date: 2018-08-10 09:13:59.383272

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '32e5c35b26a8'
down_revision = 'd3c8a090f3de'


def upgrade():
    op.add_column(u'quotas',
                  sa.Column('l7policy', sa.Integer(), nullable=True))
    op.add_column(u'quotas',
                  sa.Column('l7rule', sa.Integer(), nullable=True))
    op.add_column(u'quotas',
                  sa.Column('in_use_l7policy', sa.Integer(), nullable=True))
    op.add_column(u'quotas',
                  sa.Column('in_use_l7rule', sa.Integer(), nullable=True))



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

'''initial_create

Revision ID: 35dee79d5865
Revises: None
Create Date: 2014-08-15 11:01:14.897223

'''

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '35dee79d5865'
down_revision = None


def upgrade():
    # Create lookup tables
    op.create_table(
        u'health_monitor_type',
        sa.Column(u'name', sa.String(30), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    # Create temporary table for table data seeding
    insert_table = sql.table(
        u'health_monitor_type',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'HTTP'},
            {'name': 'HTTPS'},
            {'name': 'TCP'}
        ]
    )

    op.create_table(
        u'protocol',
        sa.Column(u'name', sa.String(30), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'protocol',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'HTTP'},
            {'name': 'HTTPS'},
            {'name': 'TCP'}
        ]
    )

    op.create_table(
        u'algorithm',
        sa.Column(u'name', sa.String(30), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'algorithm',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'ROUND_ROBIN'},
            {'name': 'LEAST_CONNECTIONS'},
            {'name': 'SOURCE_IP'}
        ]
    )

    op.create_table(
        u'session_persistence_type',
        sa.Column(u'name', sa.String(30), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'session_persistence_type',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'SOURCE_IP'},
            {'name': 'HTTP_COOKIE'},
            {'name': 'APP_COOKIE'}
        ]
    )

    op.create_table(
        u'provisioning_status',
        sa.Column(u'name', sa.String(30), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'provisioning_status',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'ACTIVE'},
            {'name': 'PENDING_CREATE'},
            {'name': 'PENDING_UPDATE'},
            {'name': 'PENDING_DELETE'},
            {'name': 'DELETED'},
            {'name': 'ERROR'}
        ]
    )

    op.create_table(
        u'operating_status',
        sa.Column(u'name', sa.String(30), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'operating_status',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'ONLINE'},
            {'name': 'OFFLINE'},
            {'name': 'DEGRADED'},
            {'name': 'ERROR'}
        ]
    )

    op.create_table(
        u'pool',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'protocol', sa.String(16), nullable=False),
        sa.Column(u'lb_algorithm', sa.String(16), nullable=False),
        sa.Column(u'operating_status', sa.String(16), nullable=False),
        sa.Column(u'enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint([u'protocol'],
                                [u'protocol.name'],
                                name=u'fk_pool_protocol_name'),
        sa.ForeignKeyConstraint([u'lb_algorithm'],
                                [u'algorithm.name'],
                                name=u'fk_pool_algorithm_name'),
        sa.ForeignKeyConstraint([u'operating_status'],
                                [u'operating_status.name'],
                                name=u'fk_pool_operating_status_name')
    )

    op.create_table(
        u'health_monitor',
        sa.Column(u'pool_id', sa.String(36), nullable=False),
        sa.Column(u'type', sa.String(36), nullable=False),
        sa.Column(u'delay', sa.Integer(), nullable=False),
        sa.Column(u'timeout', sa.Integer(), nullable=False),
        sa.Column(u'fall_threshold', sa.Integer(), nullable=False),
        sa.Column(u'rise_threshold', sa.Integer(), nullable=False),
        sa.Column(u'http_method', sa.String(16), nullable=True),
        sa.Column(u'url_path', sa.String(255), nullable=True),
        sa.Column(u'expected_codes', sa.String(64), nullable=True),
        sa.Column(u'enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'pool_id'),
        sa.ForeignKeyConstraint([u'pool_id'], [u'pool.id'],
                                name=u'fk_health_monitor_pool_id'),
        sa.ForeignKeyConstraint(
            [u'type'], [u'health_monitor_type.name'],
            name=u'fk_health_monitor_health_monitor_type_name')
    )

    op.create_table(
        u'session_persistence',
        sa.Column(u'pool_id', sa.String(36), nullable=False),
        sa.Column(u'type', sa.String(16), nullable=False),
        sa.Column(u'cookie_name', sa.String(255), nullable=True),
        sa.ForeignKeyConstraint(
            [u'type'], [u'session_persistence_type.name'],
            name=u'fk_session_persistence_session_persistence_type_name'),
        sa.ForeignKeyConstraint([u'pool_id'], [u'pool.id'],
                                name=u'fk_session_persistence_pool_id'),
        sa.PrimaryKeyConstraint(u'pool_id')
    )

    op.create_table(
        u'member',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'pool_id', sa.String(36), nullable=False),
        sa.Column(u'subnet_id', sa.String(36), nullable=True),
        sa.Column(u'address', sa.String(64), nullable=False),
        sa.Column(u'protocol_port', sa.Integer(), nullable=False),
        sa.Column(u'weight', sa.Integer(), nullable=True),
        sa.Column(u'operating_status', sa.String(16), nullable=False),
        sa.Column(u'enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint([u'pool_id'], [u'pool.id'],
                                name=u'fk_member_pool_id'),
        sa.ForeignKeyConstraint([u'operating_status'],
                                [u'operating_status.name'],
                                name=u'fk_member_operating_status_name'),
        sa.UniqueConstraint(u'pool_id', u'address', u'protocol_port',
                            name=u'uq_member_pool_id_address_protocol_port')
    )

    op.create_table(
        u'load_balancer',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'provisioning_status', sa.String(16), nullable=False),
        sa.Column(u'operating_status', sa.String(16), nullable=False),
        sa.Column(u'enabled', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint(
            [u'provisioning_status'], [u'provisioning_status.name'],
            name=u'fk_load_balancer_provisioning_status_name'),
        sa.ForeignKeyConstraint([u'operating_status'],
                                [u'operating_status.name'],
                                name=u'fk_load_balancer_operating_status_name')
    )

    op.create_table(
        u'vip',
        sa.Column(u'load_balancer_id', sa.String(36), nullable=False),
        sa.Column(u'ip_address', sa.String(36), nullable=True),
        sa.Column(u'net_port_id', sa.String(36), nullable=True),
        sa.Column(u'subnet_id', sa.String(36), nullable=True),
        sa.Column(u'floating_ip_id', sa.String(36), nullable=True),
        sa.Column(u'floating_ip_network_id', sa.String(36), nullable=True),
        sa.PrimaryKeyConstraint(u'load_balancer_id'),
        sa.ForeignKeyConstraint([u'load_balancer_id'], [u'load_balancer.id'],
                                name=u'fk_vip_load_balancer_id')
    )

    op.create_table(
        u'listener',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'protocol', sa.String(16), nullable=False),
        sa.Column(u'protocol_port', sa.Integer(), nullable=False),
        sa.Column(u'connection_limit', sa.Integer(), nullable=True),
        sa.Column(u'load_balancer_id', sa.String(36), nullable=True),
        sa.Column(u'tls_certificate_id', sa.String(36), nullable=True),
        sa.Column(u'default_pool_id', sa.String(36), nullable=True),
        sa.Column(u'provisioning_status', sa.String(16), nullable=False),
        sa.Column(u'operating_status', sa.String(16), nullable=False),
        sa.Column(u'enabled', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint([u'load_balancer_id'], [u'load_balancer.id'],
                                name=u'fk_listener_load_balancer_id'),
        sa.ForeignKeyConstraint([u'default_pool_id'], [u'pool.id'],
                                name=u'fk_listener_pool_id'),
        sa.ForeignKeyConstraint([u'protocol'], [u'protocol.name'],
                                name=u'fk_listener_protocol_name'),
        sa.ForeignKeyConstraint([u'provisioning_status'],
                                [u'provisioning_status.name'],
                                name=u'fk_listener_provisioning_status_name'),
        sa.ForeignKeyConstraint([u'operating_status'],
                                [u'operating_status.name'],
                                name=u'fk_listener_operating_status_name'),
        sa.UniqueConstraint(u'default_pool_id',
                            name=u'uq_listener_default_pool_id'),
        sa.UniqueConstraint(
            u'load_balancer_id', u'protocol_port',
            name=u'uq_listener_load_balancer_id_protocol_port'),
        sa.PrimaryKeyConstraint(u'id')
    )

    op.create_table(
        u'sni',
        sa.Column(u'listener_id', sa.String(36), nullable=False),
        sa.Column(u'tls_container_id', sa.String(36), nullable=False),
        sa.Column(u'position', sa.Integer, nullable=True),
        sa.ForeignKeyConstraint([u'listener_id'], [u'listener.id'],
                                name=u'fk_sni_listener_id'),
        sa.PrimaryKeyConstraint(u'listener_id', u'tls_container_id')
    )

    op.create_table(
        u'listener_statistics',
        sa.Column(u'listener_id', sa.String(36), nullable=False),
        sa.Column(u'bytes_in', sa.BigInteger(), nullable=False),
        sa.Column(u'bytes_out', sa.BigInteger(), nullable=False),
        sa.Column(u'active_connections', sa.Integer(), nullable=False),
        sa.Column(u'total_connections', sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint(u'listener_id'),
        sa.ForeignKeyConstraint([u'listener_id'], [u'listener.id'],
                                name=u'fk_listener_statistics_listener_id')
    )

    op.create_table(
        u'amphora',
        # id should come from the service providing the amphora (i.e. nova)
        sa.Column(u'id', sa.String(36), nullable=False, autoincrement=False),
        sa.Column(u'host_id', sa.String(36), nullable=False),
        sa.Column(u'status', sa.String(36), nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint(
            [u'status'], [u'provisioning_status.name'],
            name=u'fk_container_provisioning_status_name')
    )

    op.create_table(
        u'load_balancer_amphora',
        sa.Column(u'amphora_id', sa.String(36), nullable=False),
        sa.Column(u'load_balancer_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(
            [u'load_balancer_id'], [u'load_balancer.id'],
            name=u'fk_load_balancer_amphora_load_balancer_id'),
        sa.ForeignKeyConstraint([u'amphora_id'],
                                [u'amphora.id'],
                                name=u'fk_load_balancer_amphora_id'),
        sa.PrimaryKeyConstraint(u'amphora_id', u'load_balancer_id')
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_utils import uuidutils

from octavia.amphorae.backends.agent.api_server import util
from octavia.amphorae.backends.utils import keepalivedlvs_query as lvs_query
from octavia.common import constants
from octavia.tests.common import utils as test_utils
from octavia.tests.unit import base

# Kernal_file_sample which is in /proc/net/ip_vs
# The realservers and the listened ports are
# 10.0.0.25:2222, 10.0.0.35:3333.
# Realserver 10.0.0.45:4444 is not listed because healthcheck failed.
# The virtual server and the listened port is
# 10.0.0.37:7777.
KERNAL_FILE_SAMPLE_V4 = (
    "IP Virtual Server version 1.2.1 (size=4096)\n"
    "Prot LocalAddress:Port Scheduler Flags\n"
    "  -> RemoteAddress:Port Forward Weight ActiveConn InActConn\n"
    "UDP  0A000025:1E61 rr\n"
    "  -> 0A000023:0D05      Masq    2      0          0\n"
    "  -> 0A000019:08AE      Masq    3      0          0")
# Kernal_file_sample which is in /proc/net/ip_vs
# The realservers and the listened ports are
# [fd79:35e2:9963:0:f816:3eff:feca:b7bf]:2222,
# [fd79:35e2:9963:0:f816:3eff:fe9d:94df]:3333.
# Readlserver [fd79:35e2:9963:0:f816:3eff:fe9d:8f3f]:4444 is not listed
# because healthcheck failed.
# The virtual server and the listened port is
# [fd79:35e2:9963:0:f816:3eff:fe6d:7a2a]:7777.
KERNAL_FILE_SAMPLE_V6 = (
    "IP Virtual Server version 1.2.1 (size=4096)\n"
    "Prot LocalAddress:Port Scheduler Flags\n"
    "  -> RemoteAddress:Port Forward Weight ActiveConn InActConn\n"
    "UDP  [fd79:35e2:9963:0000:f816:3eff:fe6d:7a2a]:1E61 rr\n"
    "  -> [fd79:35e2:9963:0000:f816:3eff:feca:b7bf]:08AE      "
    "Masq    3      0          0\n"
    "  -> [fd79:35e2:9963:0000:f816:3eff:fe9d:94df]:0D05      "
    "Masq    2      0          0\n"
    "  -> [fd79:35e2::8f3f]:115C       "
    "Masq    2      0          0")

CFG_FILE_TEMPLATE_v4 = (
    "# Configuration for Listener %(listener_id)s\n\n"
    "net_namespace %(ns_name)s\n\n"
    "virtual_server 10.0.0.37 7777 {\n"
    "    lb_algo rr\n"
    "    lb_kind NAT\n"
    "    protocol udp\n\n\n"
    "    # Configuration for Pool %(pool_id)s\n"
    "    # Configuration for Member %(member_id1)s\n"
    "    real_server 10.0.0.25 2222 {\n"
    "        weight 3\n"
    "        persistence_timeout 5\n"
    "        persistence_granularity 255.0.0.0\n\n"
    "        MISC_CHECK {\n\n"
    "            misc_path \"/usr/bin/check_script.sh\"\n\n"
    "            misc_timeout 5\n\n"
    "        }\n\n"
    "    }\n\n"
    "    # Configuration for Member %(member_id2)s\n"
    "    real_server 10.0.0.35 3333 {\n"
    "        weight 2\n"
    "        persistence_timeout 5\n"
    "        persistence_granularity 255.0.0.0\n\n"
    "        MISC_CHECK {\n\n"
    "            misc_path \"/usr/bin/check_script.sh\"\n\n"
    "            misc_timeout 5\n\n"
    "        }\n\n"
    "    }\n\n"
    "    # Configuration for Member %(member_id3)s\n"
    "    real_server 10.0.0.45 4444 {\n"
    "        weight 2\n"
    "        persistence_timeout 5\n"
    "        persistence_granularity 255.0.0.0\n\n"
    "        MISC_CHECK {\n\n"
    "            misc_path \"/usr/bin/check_script.sh\"\n\n"
    "            misc_timeout 5\n\n"
    "        }\n\n"
    "    }\n\n"
    "    # Member %(member_id4)s is disabled\n\n"
    "}")

CFG_FILE_TEMPLATE_v6 = (
    "# Configuration for Listener %(listener_id)s\n\n"
    "net_namespace %(ns_name)s\n\n"
    "virtual_server fd79:35e2:9963:0:f816:3eff:fe6d:7a2a 7777 {\n"
    "    lb_algo rr\n"
    "    lb_kind NAT\n"
    "    protocol udp\n\n\n"
    "    # Configuration for Pool %(pool_id)s\n"
    "    # Configuration for Member %(member_id1)s\n"
    "    real_server fd79:35e2:9963:0:f816:3eff:feca:b7bf 2222 {\n"
    "        weight 3\n"
    "        MISC_CHECK {\n\n"
    "            misc_path \"/usr/bin/check_script.sh\"\n\n"
    "            misc_timeout 5\n\n"
    "        }\n\n"
    "    }\n\n"
    "    # Configuration for Member %(member_id2)s\n"
    "    real_server fd79:35e2:9963:0:f816:3eff:fe9d:94df 3333 {\n"
    "        weight 2\n"
    "        MISC_CHECK {\n\n"
    "            misc_path \"/usr/bin/check_script.sh\"\n\n"
    "            misc_timeout 5\n\n"
    "        }\n\n"
    "    }\n\n"
    "    # Configuration for Member %(member_id3)s\n"
    "    real_server fd79:35e2:9963:0:f816:3eff:fe9d:8f3f 4444 {\n"
    "        weight 2\n"
    "        MISC_CHECK {\n\n"
    "            misc_path \"/usr/bin/check_script.sh\"\n\n"
    "            misc_timeout 5\n\n"
    "        }\n\n"
    "    }\n\n"
    "    # Member %(member_id4)s is disabled\n\n"
    "    # Configuration for Member %(member_id5)s\n"
    "    real_server fd79:35e2:0:0:0:0:0:8f3f 4444 {\n"
    "        weight 2\n"
    "        MISC_CHECK {\n\n"
    "            misc_path \"/usr/bin/check_script.sh\"\n\n"
    "            misc_timeout 5\n\n"
    "        }\n\n"
    "    }\n\n"
    "}")

CFG_FILE_TEMPLATE_DISABLED_LISTENER = (
    "# Listener %(listener_id)s is disabled \n\n"
    "net_namespace %(ns_name)s\n\n"
)

IPVSADM_OUTPUT_TEMPLATE = (
    "IP Virtual Server version 1.2.1 (size=4096)\n"
    "Prot LocalAddress:Port Scheduler Flags\n"
    "  -> RemoteAddress:Port           Forward Weight ActiveConn InActConn\n"
    "UDP  %(listener_ipport)s rr\n"
    "  -> %(member1_ipport)s               Masq    3      0          0\n"
    "  -> %(member2_ipport)s               Masq    2      0          0")

IPVSADM_STATS_OUTPUT_TEMPLATE = (
    "IP Virtual Server version 1.2.1 (size=4096)\n"
    "Prot LocalAddress:Port               Conns   InPkts  OutPkts  "
    "InBytes OutBytes\n"
    "  -> RemoteAddress:Port\n"
    "UDP  %(listener_ipport)s                      5     4264        5"
    "  6387472     7490\n"
    "  -> %(member1_ipport)s                      2     1706        2"
    "  2555588     2996\n"
    "  -> %(member2_ipport)s                      3     2558        3"
    "  3831884     4494")


class LvsQueryTestCase(base.TestCase):
    def setUp(self):
        super().setUp()
        self.listener_id_v4 = uuidutils.generate_uuid()
        self.pool_id_v4 = uuidutils.generate_uuid()
        self.member_id1_v4 = uuidutils.generate_uuid()
        self.member_id2_v4 = uuidutils.generate_uuid()
        self.member_id3_v4 = uuidutils.generate_uuid()
        self.member_id4_v4 = uuidutils.generate_uuid()
        self.listener_id_v6 = uuidutils.generate_uuid()
        self.pool_id_v6 = uuidutils.generate_uuid()
        self.member_id1_v6 = uuidutils.generate_uuid()
        self.member_id2_v6 = uuidutils.generate_uuid()
        self.member_id3_v6 = uuidutils.generate_uuid()
        self.member_id4_v6 = uuidutils.generate_uuid()
        self.member_id5_v6 = uuidutils.generate_uuid()
        self.disabled_listener_id = uuidutils.generate_uuid()
        cfg_content_v4 = CFG_FILE_TEMPLATE_v4 % {
            'listener_id': self.listener_id_v4,
            'ns_name': constants.AMPHORA_NAMESPACE,
            'pool_id': self.pool_id_v4,
            'member_id1': self.member_id1_v4,
            'member_id2': self.member_id2_v4,
            'member_id3': self.member_id3_v4,
            'member_id4': self.member_id4_v4,
        }
        cfg_content_v6 = CFG_FILE_TEMPLATE_v6 % {
            'listener_id': self.listener_id_v6,
            'ns_name': constants.AMPHORA_NAMESPACE,
            'pool_id': self.pool_id_v6,
            'member_id1': self.member_id1_v6,
            'member_id2': self.member_id2_v6,
            'member_id3': self.member_id3_v6,
            'member_id4': self.member_id4_v6,
            'member_id5': self.member_id5_v6
        }
        cfg_content_disabled_listener = (
            CFG_FILE_TEMPLATE_DISABLED_LISTENER % {
                'listener_id': self.listener_id_v6,
                'ns_name': constants.AMPHORA_NAMESPACE,
            }
        )
        self.useFixture(test_utils.OpenFixture(
            util.keepalived_lvs_cfg_path(self.listener_id_v4), cfg_content_v4))
        self.useFixture(test_utils.OpenFixture(
            util.keepalived_lvs_cfg_path(self.listener_id_v6), cfg_content_v6))
        self.useFixture(test_utils.OpenFixture(
            util.keepalived_lvs_cfg_path(self.disabled_listener_id),
            cfg_content_disabled_listener))

    @mock.patch('subprocess.check_output')
    def test_get_listener_realserver_mapping(self, mock_check_output):
        # Ipv4 resolver
        input_listener_ip_port = '10.0.0.37:7777'
        target_ns = constants.AMPHORA_NAMESPACE
        mock_check_output.return_value = KERNAL_FILE_SAMPLE_V4
        result = lvs_query.get_listener_realserver_mapping(
            target_ns, input_listener_ip_port,
            health_monitor_enabled=True)
        expected = {'10.0.0.25:2222': {'status': 'UP',
                                       'Forward': 'Masq',
                                       'Weight': '3',
                                       'ActiveConn': '0',
                                       'InActConn': '0'},
                    '10.0.0.35:3333': {'status': 'UP',
                                       'Forward': 'Masq',
                                       'Weight': '2',
                                       'ActiveConn': '0',
                                       'InActConn': '0'}}
        self.assertEqual((True, expected), result)

        # Ipv6 resolver
        input_listener_ip_port = '[fd79:35e2:9963:0:f816:3eff:fe6d:7a2a]:7777'
        mock_check_output.return_value = KERNAL_FILE_SAMPLE_V6
        result = lvs_query.get_listener_realserver_mapping(
            target_ns, input_listener_ip_port,
            health_monitor_enabled=True)
        expected = {'[fd79:35e2:9963:0:f816:3eff:feca:b7bf]:2222':
                    {'status': constants.UP,
                     'Forward': 'Masq',
                     'Weight': '3',
                     'ActiveConn': '0',
                     'InActConn': '0'},
                    '[fd79:35e2:9963:0:f816:3eff:fe9d:94df]:3333':
                        {'status': constants.UP,
                         'Forward': 'Masq',
                         'Weight': '2',
                         'ActiveConn': '0',
                         'InActConn': '0'},
                    '[fd79:35e2::8f3f]:4444':
                        {'status': constants.UP,
                         'Forward': 'Masq',
                         'Weight': '2',
                         'ActiveConn': '0',
                         'InActConn': '0'}}
        self.assertEqual((True, expected), result)

        # negetive cases
        mock_check_output.return_value = KERNAL_FILE_SAMPLE_V4
        for listener_ip_port in ['10.0.0.37:7776', '10.0.0.31:7777']:
            result = lvs_query.get_listener_realserver_mapping(
                target_ns, listener_ip_port,
                health_monitor_enabled=True)
            self.assertEqual((False, {}), result)

        mock_check_output.return_value = KERNAL_FILE_SAMPLE_V6
        for listener_ip_port in [
            '[fd79:35e2:9963:0:f816:3eff:fe6d:7a2a]:7776',
                '[fd79:35e2:9973:0:f816:3eff:fe6d:7a2a]:7777']:
            result = lvs_query.get_listener_realserver_mapping(
                target_ns, listener_ip_port,
                health_monitor_enabled=True)
            self.assertEqual((False, {}), result)

    def test_get_lvs_listener_resource_ipports_nsname(self):
        # ipv4
        res = lvs_query.get_lvs_listener_resource_ipports_nsname(
            self.listener_id_v4)
        expected = {'Listener': {'id': self.listener_id_v4,
                                 'ipport': '10.0.0.37:7777'},
                    'Pool': {'id': self.pool_id_v4},
                    'Members': [{'id': self.member_id1_v4,
                                 'ipport': '10.0.0.25:2222'},
                                {'id': self.member_id2_v4,
                                 'ipport': '10.0.0.35:3333'},
                                {'id': self.member_id3_v4,
                                 'ipport': '10.0.0.45:4444'},
                                {'id': self.member_id4_v4,
                                 'ipport': None}]}
        self.assertEqual((expected, constants.AMPHORA_NAMESPACE), res)

        # ipv6
        res = lvs_query.get_lvs_listener_resource_ipports_nsname(
            self.listener_id_v6)
        expected = {'Listener': {
            'id': self.listener_id_v6,
            'ipport': '[fd79:35e2:9963:0:f816:3eff:fe6d:7a2a]:7777'},
            'Pool': {'id': self.pool_id_v6},
            'Members': [
                {'id': self.member_id1_v6,
                 'ipport': '[fd79:35e2:9963:0:f816:3eff:feca:b7bf]:2222'},
                {'id': self.member_id2_v6,
                 'ipport': '[fd79:35e2:9963:0:f816:3eff:fe9d:94df]:3333'},
                {'id': self.member_id3_v6,
                 'ipport': '[fd79:35e2:9963:0:f816:3eff:fe9d:8f3f]:4444'},
                {'id': self.member_id5_v6,
                 'ipport': '[fd79:35e2::8f3f]:4444'},
                {'id': self.member_id4_v6,
                 'ipport': None}]}
        self.assertEqual((expected, constants.AMPHORA_NAMESPACE), res)

        # disabled
        res = lvs_query.get_lvs_listener_resource_ipports_nsname(
            self.disabled_listener_id)
        self.assertEqual((None, constants.AMPHORA_NAMESPACE), res)

    @mock.patch('os.stat')
    @mock.patch('subprocess.check_output')
    def test_get_lvs_listener_pool_status(self, mock_check_output,
                                          mock_os_stat):
        mock_os_stat.side_effect = (
            mock.Mock(st_mtime=1234),
            mock.Mock(st_mtime=1234),
        )

        # test with ipv4 and ipv6
        mock_check_output.return_value = KERNAL_FILE_SAMPLE_V4
        res = lvs_query.get_lvs_listener_pool_status(self.listener_id_v4)
        expected = {
            'lvs':
            {'uuid': self.pool_id_v4,
             'status': constants.UP,
             'members': {self.member_id1_v4: constants.UP,
                         self.member_id2_v4: constants.UP,
                         self.member_id3_v4: constants.DOWN,
                         self.member_id4_v4: constants.MAINT}}}
        self.assertEqual(expected, res)

        mock_os_stat.side_effect = (
            mock.Mock(st_mtime=1234),
            mock.Mock(st_mtime=1234),
        )

        mock_check_output.return_value = KERNAL_FILE_SAMPLE_V6
        res = lvs_query.get_lvs_listener_pool_status(self.listener_id_v6)
        expected = {
            'lvs':
            {'uuid': self.pool_id_v6,
             'status': constants.UP,
             'members': {self.member_id1_v6: constants.UP,
                         self.member_id2_v6: constants.UP,
                         self.member_id3_v6: constants.DOWN,
                         self.member_id4_v6: constants.MAINT,
                         self.member_id5_v6: constants.UP}}}
        self.assertEqual(expected, res)

    @mock.patch('os.stat')
    @mock.patch('subprocess.check_output')
    def test_get_lvs_listener_pool_status_restarting(self, mock_check_output,
                                                     mock_os_stat):
        mock_os_stat.side_effect = (
            mock.Mock(st_mtime=1234),  # config file
            mock.Mock(st_mtime=1220),  # pid file
        )

        # test with ipv4 and ipv6
        mock_check_output.return_value = KERNAL_FILE_SAMPLE_V4
        res = lvs_query.get_lvs_listener_pool_status(self.listener_id_v4)
        expected = {
            'lvs':
            {'uuid': self.pool_id_v4,
             'status': constants.UP,
             'members': {self.member_id1_v4: constants.UP,
                         self.member_id2_v4: constants.UP,
                         self.member_id3_v4: constants.RESTARTING,
                         self.member_id4_v4: constants.MAINT}}}
        self.assertEqual(expected, res)

    @mock.patch('octavia.amphorae.backends.utils.keepalivedlvs_query.'
                'get_lvs_listener_resource_ipports_nsname')
    def test_get_lvs_listener_pool_status_when_no_pool(
            self, mock_get_resource_ipports):
        # Just test with ipv4, ipv6 tests is same.
        # the returned resource_ipport_mapping doesn't contains the 'Pool'
        # resource, that means the listener doesn't have a pool resource, it
        # isn't usable at this moment, then the pool status will
        # return nothing.
        mock_get_resource_ipports.return_value = (
            {
                'Listener': {
                    'id': self.listener_id_v4,
                    'ipport': '10.0.0.37:7777'}},
            constants.AMPHORA_NAMESPACE)
        res = lvs_query.get_lvs_listener_pool_status(self.listener_id_v4)
        self.assertEqual({}, res)

    @mock.patch('octavia.amphorae.backends.utils.keepalivedlvs_query.'
                'get_lvs_listener_resource_ipports_nsname')
    def test_get_lvs_listener_pool_status_when_no_members(
            self, mock_get_resource_ipports):
        # Just test with ipv4, ipv6 tests is same.
        # the returned resource_ipport_mapping doesn't contains the 'Members'
        # resources, that means the pool of listener doesn't have a enabled
        # pool resource, so the pool is not usable, then the pool status will
        # return UP.
        mock_get_resource_ipports.return_value = (
            {
                'Listener': {'id': self.listener_id_v4,
                             'ipport': '10.0.0.37:7777'},
                'Pool': {'id': self.pool_id_v4}},
            constants.AMPHORA_NAMESPACE)
        res = lvs_query.get_lvs_listener_pool_status(self.listener_id_v4)
        expected = {'lvs': {
            'uuid': self.pool_id_v4,
            'status': constants.UP,
            'members': {}
        }}
        self.assertEqual(expected, res)

    @mock.patch('os.stat')
    @mock.patch('octavia.amphorae.backends.utils.keepalivedlvs_query.'
                'get_listener_realserver_mapping')
    def test_get_lvs_listener_pool_status_when_not_get_realserver_result(
            self, mock_get_mapping, mock_os_stat):
        # This will hit if the kernel lvs file (/proc/net/ip_vs)
        # lose its content. So at this moment, eventhough we configure the
        # pool and member into udp keepalived config file, we have to set
        # ths status of pool and its members to DOWN.
        mock_os_stat.side_effect = (
            mock.Mock(st_mtime=1234),
            mock.Mock(st_mtime=1234),
        )
        mock_get_mapping.return_value = (False, {})
        res = lvs_query.get_lvs_listener_pool_status(self.listener_id_v4)
        expected = {
            'lvs':
            {'uuid': self.pool_id_v4,
             'status': constants.DOWN,
             'members': {self.member_id1_v4: constants.DOWN,
                         self.member_id2_v4: constants.DOWN,
                         self.member_id3_v4: constants.DOWN,
                         self.member_id4_v4: constants.MAINT}}}
        self.assertEqual(expected, res)

    @mock.patch('subprocess.check_output')
    def test_get_ipvsadm_info(self, mock_check_output):
        for ip_list in [["10.0.0.37:7777", "10.0.0.25:2222", "10.0.0.35:3333"],
                        ["[fd79:35e2:9963:0:f816:3eff:fe6d:7a2a]:7777",
                         "[fd79:35e2:9963:0:f816:3eff:feca:b7bf]:2222",
                         "[fd79:35e2:9963:0:f816:3eff:fe9d:94df]:3333"]]:
            mock_check_output.return_value = IPVSADM_OUTPUT_TEMPLATE % {
                "listener_ipport": ip_list[0],
                "member1_ipport": ip_list[1],
                "member2_ipport": ip_list[2]}
            res = lvs_query.get_ipvsadm_info(constants.AMPHORA_NAMESPACE)
            # This expected result can reference on IPVSADM_OUTPUT_TEMPLATE,
            # that means the function can get every element of the virtual
            # server and the real servers.
            expected = {
                ip_list[0]: {
                    'Listener': [('Prot', 'UDP'),
                                 ('LocalAddress:Port', ip_list[0]),
                                 ('Scheduler', 'rr')],
                    'Members': [[('RemoteAddress:Port', ip_list[1]),
                                 ('Forward', 'Masq'), ('Weight', '3'),
                                 ('ActiveConn', '0'), ('InActConn', '0')],
                                [('RemoteAddress:Port', ip_list[2]),
                                 ('Forward', 'Masq'), ('Weight', '2'),
                                 ('ActiveConn', '0'), ('InActConn', '0')]]}}
            self.assertEqual(expected, res)

            # ipvsadm stats
            mock_check_output.return_value = IPVSADM_STATS_OUTPUT_TEMPLATE % {
                "listener_ipport": ip_list[0],
                "member1_ipport": ip_list[1],
                "member2_ipport": ip_list[2]}
            res = lvs_query.get_ipvsadm_info(constants.AMPHORA_NAMESPACE,
                                             is_stats_cmd=True)
            expected = {
                ip_list[0]:
                    {'Listener': [('Prot', 'UDP'),
                                  ('LocalAddress:Port', ip_list[0]),
                                  ('Conns', '5'),
                                  ('InPkts', '4264'),
                                  ('OutPkts', '5'),
                                  ('InBytes', '6387472'),
                                  ('OutBytes', '7490')],
                     'Members': [[('RemoteAddress:Port', ip_list[1]),
                                  ('Conns', '2'),
                                  ('InPkts', '1706'),
                                  ('OutPkts', '2'),
                                  ('InBytes', '2555588'),
                                  ('OutBytes', '2996')],
                                 [('RemoteAddress:Port', ip_list[2]),
                                  ('Conns', '3'),
                                  ('InPkts', '2558'),
                                  ('OutPkts', '3'),
                                  ('InBytes', '3831884'),
                                  ('OutBytes', '4494')]]}}
            self.assertEqual(expected, res)

    @mock.patch('subprocess.check_output')
    @mock.patch("octavia.amphorae.backends.agent.api_server.util."
                "is_lvs_listener_running", return_value=True)
    @mock.patch("octavia.amphorae.backends.agent.api_server.util."
                "get_lvs_listeners")
    def test_get_lvs_listeners_stats(
            self, mock_get_listener, mock_is_running, mock_check_output):
        # The ipv6 test is same with ipv4, so just test ipv4 here
        mock_get_listener.return_value = [self.listener_id_v4]
        output_list = list()
        output_list.append(IPVSADM_OUTPUT_TEMPLATE % {
            "listener_ipport": "10.0.0.37:7777",
            "member1_ipport": "10.0.0.25:2222",
            "member2_ipport": "10.0.0.35:3333"})
        output_list.append(IPVSADM_STATS_OUTPUT_TEMPLATE % {
            "listener_ipport": "10.0.0.37:7777",
            "member1_ipport": "10.0.0.25:2222",
            "member2_ipport": "10.0.0.35:3333"})
        mock_check_output.side_effect = output_list
        res = lvs_query.get_lvs_listeners_stats()
        # We can check the expected result reference the stats sample,
        # that means this func can compute the stats info of single listener.
        expected = {self.listener_id_v4: {
            'status': constants.OPEN,
            'stats': {'bin': 6387472, 'stot': 5, 'bout': 7490,
                      'ereq': 0, 'scur': 0}}}
        self.assertEqual(expected, res)

        # if no udp listener need to be collected.
        # Then this function will return nothing.
        mock_is_running.return_value = False
        res = lvs_query.get_lvs_listeners_stats()
        self.assertEqual({}, res)

    @mock.patch('subprocess.check_output')
    @mock.patch("octavia.amphorae.backends.agent.api_server.util."
                "is_lvs_listener_running", return_value=True)
    @mock.patch("octavia.amphorae.backends.agent.api_server.util."
                "get_lvs_listeners")
    def test_get_lvs_listeners_stats_disabled_listener(
            self, mock_get_listener, mock_is_running, mock_check_output):
        mock_get_listener.return_value = [self.disabled_listener_id]
        res = lvs_query.get_lvs_listeners_stats()
        self.assertEqual({}, res)



# Copyright 2011 VMware, Inc., 2014 A10 Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Octavia
"""

import os
import ssl
import sys

from keystoneauth1 import loading as ks_loading
from octavia_lib.common import constants as lib_consts
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_policy import opts as policy_opts

from octavia.certificates.common import local
from octavia.common import constants
from octavia.common import utils
from octavia.common import validate
from octavia.i18n import _
from octavia import version

LOG = logging.getLogger(__name__)

EXTRA_LOG_LEVEL_DEFAULTS = [
    'neutronclient.v2_0.client=INFO',
]

TLS_PROTOCOL_CHOICES = [
    p[9:].replace('_', '.') for p in ssl._PROTOCOL_NAMES.values()]

core_opts = [
    cfg.HostnameOpt('host', default=utils.get_hostname(),
                    sample_default='<server-hostname.example.com>',
                    help=_("The hostname Octavia is running on")),
    cfg.StrOpt('octavia_plugins', default='hot_plug_plugin',
               help=_("Name of the controller plugin to use")),
]

api_opts = [
    cfg.IPOpt('bind_host', default='127.0.0.1',
              help=_("The host IP to bind to")),
    cfg.PortOpt('bind_port', default=9876,
                help=_("The port to bind to")),
    cfg.StrOpt('auth_strategy', default=constants.KEYSTONE,
               choices=[constants.NOAUTH,
                        constants.KEYSTONE,
                        constants.TESTING],
               help=_("The auth strategy for API requests.")),
    cfg.BoolOpt('allow_pagination', default=True,
                help=_("Allow the usage of pagination")),
    cfg.BoolOpt('allow_sorting', default=True,
                help=_("Allow the usage of sorting")),
    cfg.BoolOpt('allow_filtering', default=True,
                help=_("Allow the usage of filtering")),
    cfg.BoolOpt('allow_field_selection', default=True,
                help=_("Allow the usage of field selection")),
    cfg.StrOpt('pagination_max_limit',
               default=str(constants.DEFAULT_PAGE_SIZE),
               help=_("The maximum number of items returned in a single "
                      "response. The string 'infinite' or a negative "
                      "integer value means 'no limit'")),
    cfg.StrOpt('api_base_uri',
               help=_("Base URI for the API for use in pagination links. "
                      "This will be autodetected from the request if not "
                      "overridden here.")),
    cfg.BoolOpt('allow_tls_terminated_listeners', default=True,
                help=_("Allow users to create TLS Terminated listeners?")),
    cfg.BoolOpt('allow_ping_health_monitors', default=True,
                help=_("Allow users to create PING type Health Monitors?")),
    cfg.DictOpt('enabled_provider_drivers',
                help=_('A comma separated list of dictionaries of the '
                       'enabled provider driver names and descriptions. '
                       'Must match the driver name in the '
                       'octavia.api.drivers entrypoint. Example: '
                       'amphora:The Octavia Amphora driver.,'
                       'octavia:Deprecated alias of the Octavia '
                       'Amphora driver.'),
                default={'amphora': 'The Octavia Amphora driver.',
                         'octavia': 'Deprecated alias of the Octavia Amphora '
                         'driver.',
                         }),
    cfg.StrOpt('default_provider_driver', default='amphora',
               help=_('Default provider driver.')),
    cfg.IntOpt('udp_connect_min_interval_health_monitor',
               default=3,
               help=_("The minimum health monitor delay interval for the "
                      "UDP-CONNECT Health Monitor type. A negative integer "
                      "value means 'no limit'.")),
    cfg.BoolOpt('healthcheck_enabled', default=False,
                help=_("When True, the oslo middleware healthcheck endpoint "
                       "is enabled in the Octavia API.")),
    cfg.IntOpt('healthcheck_refresh_interval', default=5,
               help=_("The interval healthcheck plugins should cache results, "
                      "in seconds.")),
    cfg.StrOpt('default_listener_ciphers',
               default=constants.CIPHERS_OWASP_SUITE_B,
               help=_("Default OpenSSL cipher string (colon-separated) for "
                      "new TLS-enabled listeners.")),
    cfg.StrOpt('default_pool_ciphers',
               default=constants.CIPHERS_OWASP_SUITE_B,
               help=_("Default OpenSSL cipher string (colon-separated) for "
                      "new TLS-enabled pools.")),
    cfg.StrOpt('tls_cipher_prohibit_list', default='',
               deprecated_name='tls_cipher_blacklist',
               help=_("Colon separated list of OpenSSL ciphers. "
                      "Usage of these ciphers will be blocked.")),
    cfg.ListOpt('default_listener_tls_versions',
                default=constants.TLS_VERSIONS_OWASP_SUITE_B,
                help=_('List of TLS versions to use for new TLS-enabled '
                       'listeners.')),
    cfg.ListOpt('default_pool_tls_versions',
                default=constants.TLS_VERSIONS_OWASP_SUITE_B,
                help=_('List of TLS versions to use for new TLS-enabled '
                       'pools.')),
    cfg.StrOpt('minimum_tls_version',
               default=None,
               choices=constants.TLS_ALL_VERSIONS + [None],
               help=_('Minimum allowed TLS version for listeners and pools.')),
    cfg.ListOpt('default_listener_alpn_protocols',
                default=[lib_consts.ALPN_PROTOCOL_HTTP_2,
                         lib_consts.ALPN_PROTOCOL_HTTP_1_1,
                         lib_consts.ALPN_PROTOCOL_HTTP_1_0],
                help=_('List of ALPN protocols to use for new TLS-enabled '
                       'listeners.')),
    cfg.ListOpt('default_pool_alpn_protocols',
                default=[lib_consts.ALPN_PROTOCOL_HTTP_2,
                         lib_consts.ALPN_PROTOCOL_HTTP_1_1,
                         lib_consts.ALPN_PROTOCOL_HTTP_1_0],
                help=_('List of ALPN protocols to use for new TLS-enabled '
                       'pools.')),
]

# Options only used by the amphora agent
amphora_agent_opts = [
    cfg.StrOpt('agent_server_ca', default='/etc/octavia/certs/client_ca.pem',
               help=_("The ca which signed the client certificates")),
    cfg.StrOpt('agent_server_cert', default='/etc/octavia/certs/server.pem',
               help=_("The server certificate for the agent server "
                      "to use")),
    cfg.StrOpt('agent_server_network_dir',
               help=_("The directory where new network interfaces "
                      "are located")),
    cfg.StrOpt('agent_server_network_file',
               help=_("The file where the network interfaces are located. "
                      "Specifying this will override any value set for "
                      "agent_server_network_dir.")),
    cfg.IntOpt('agent_request_read_timeout', default=180,
               help=_("The time in seconds to allow a request from the "
                      "controller to run before terminating the socket.")),
    cfg.StrOpt('agent_tls_protocol', default='TLSv1.2',
               help=_("Minimum TLS protocol for communication with the "
                      "amphora agent."),
               choices=TLS_PROTOCOL_CHOICES),

    # Logging setup
    cfg.ListOpt('admin_log_targets',
                help=_('List of log server ip and port pairs for '
                       'Administrative logs. Additional hosts are backup to '
                       'the primary server. If none is '
                       'specified remote logging is disabled. Example '
                       '127.0.0.1:10514, 192.168.0.1:10514')),
    cfg.ListOpt('tenant_log_targets',
                help=_('List of log server ip and port pairs for '
                       'tenant traffic logs. Additional hosts are backup to '
                       'the primary server. If none is '
                       'specified remote logging is disabled. Example '
                       '127.0.0.1:10514, 192.168.0.1:10514')),
    cfg.IntOpt('user_log_facility', default=0, min=0, max=7,
               help=_('LOG_LOCAL facility number to use for user traffic '
                      'logs.')),
    cfg.IntOpt('administrative_log_facility', default=1, min=0, max=7,
               help=_('LOG_LOCAL facility number to use for amphora processes '
                      'logs.')),
    cfg.StrOpt('log_protocol', default=lib_consts.PROTOCOL_UDP,
               choices=[lib_consts.PROTOCOL_TCP, lib_consts.PROTOCOL_UDP],
               help=_("The log forwarding transport protocol. One of UDP or "
                      "TCP.")),
    cfg.IntOpt('log_retry_count', default=5,
               help=_('The maximum attempts to retry connecting to the '
                      'logging host.')),
    cfg.IntOpt('log_retry_interval', default=2,
               help=_('The time, in seconds, to wait between retries '
                      'connecting to the logging host.')),
    cfg.IntOpt('log_queue_size', default=10000,
               help=_('The queue size (messages) to buffer log messages.')),
    cfg.StrOpt('logging_template_override',
               help=_('Custom logging configuration template.')),
    cfg.BoolOpt('forward_all_logs', default=False,
                help=_('When True, the amphora will forward all of the '
                       'system logs (except tenant traffic logs) to the '
                       'admin log target(s). When False, '
                       'only amphora specific admin logs will be forwarded.')),
    cfg.BoolOpt('disable_local_log_storage', default=False,
                help=_('When True, no logs will be written to the amphora '
                       'filesystem. When False, log files will be written to '
                       'the local filesystem.')),

    # Do not specify in octavia.conf, loaded at runtime
    cfg.StrOpt('amphora_id', help=_("The amphora ID.")),
    cfg.StrOpt('amphora_udp_driver',
               default='keepalived_lvs',
               help='The UDP API backend for amphora agent.',
               deprecated_for_removal=True,
               deprecated_reason=_('amphora-agent will not support any other '
                                   'backend than keepalived_lvs.'),
               deprecated_since='Wallaby'),
]

compute_opts = [
    cfg.IntOpt('max_retries', default=15,
               help=_('The maximum attempts to retry an action with the '
                      'compute service.')),
    cfg.IntOpt('retry_interval', default=1,
               help=_('Seconds to wait before retrying an action with the '
                      'compute service.')),
    cfg.IntOpt('retry_backoff', default=1,
               help=_('The seconds to backoff retry attempts.')),
    cfg.IntOpt('retry_max', default=10,
               help=_('The maximum interval in seconds between retry '
                      'attempts.')),
]

networking_opts = [
    cfg.IntOpt('max_retries', default=15,
               help=_('The maximum attempts to retry an action with the '
                      'networking service.')),
    cfg.IntOpt('retry_interval', default=1,
               help=_('Seconds to wait before retrying an action with the '
                      'networking service.')),
    cfg.IntOpt('retry_backoff', default=1,
               help=_('The seconds to backoff retry attempts.')),
    cfg.IntOpt('retry_max', default=10,
               help=_('The maximum interval in seconds between retry '
                      'attempts.')),
    cfg.IntOpt('port_detach_timeout', default=300,
               help=_('Seconds to wait for a port to detach from an '
                      'amphora.')),
    cfg.BoolOpt('allow_vip_network_id', default=True,
                help=_('Can users supply a network_id for their VIP?')),
    cfg.BoolOpt('allow_vip_subnet_id', default=True,
                help=_('Can users supply a subnet_id for their VIP?')),
    cfg.BoolOpt('allow_vip_port_id', default=True,
                help=_('Can users supply a port_id for their VIP?')),
    cfg.ListOpt('valid_vip_networks',
                help=_('List of network_ids that are valid for VIP '
                       'creation. If this field is empty, no validation '
                       'is performed.')),
    cfg.ListOpt('reserved_ips',
                default=['169.254.169.254'],
                item_type=cfg.types.IPAddress(),
                help=_('List of IP addresses reserved from being used for '
                       'member addresses. IPv6 addresses should be in '
                       'expanded, uppercase form.')),
    cfg.BoolOpt('allow_invisible_resource_usage', default=False,
                help=_("When True, users can use network resources they "
                       "cannot normally see as VIP or member subnets. Making "
                       "this True may allow users to access resources on "
                       "subnets they do not normally have access to via "
                       "neutron RBAC policies.")),
]

health_manager_opts = [
    cfg.IPOpt('bind_ip', default='127.0.0.1',
              help=_('IP address the controller will listen on for '
                     'heart beats')),
    cfg.PortOpt('bind_port', default=5555,
                help=_('Port number the controller will listen on '
                       'for heart beats')),
    cfg.IntOpt('failover_threads',
               default=10,
               help=_('Number of threads performing amphora failovers.')),
    cfg.IntOpt('health_update_threads',
               default=None,
               help=_('Number of processes for amphora health update.')),
    cfg.IntOpt('stats_update_threads',
               default=None,
               help=_('Number of processes for amphora stats update.')),
    cfg.StrOpt('heartbeat_key',
               mutable=True,
               help=_('key used to validate amphora sending '
                      'the message'), secret=True),
    cfg.IntOpt('heartbeat_timeout',
               default=60,
               help=_('Interval, in seconds, to wait before failing over an '
                      'amphora.')),
    cfg.IntOpt('health_check_interval',
               default=3,
               help=_('Sleep time between health checks in seconds.')),
    cfg.IntOpt('sock_rlimit', default=0,
               help=_(' sets the value of the heartbeat recv buffer')),

    # Used by the health manager on the amphora
    cfg.ListOpt('controller_ip_port_list',
                help=_('List of controller ip and port pairs for the '
                       'heartbeat receivers. Example 127.0.0.1:5555, '
                       '192.168.0.1:5555'),
                mutable=True,
                default=[]),
    cfg.IntOpt('heartbeat_interval',
               default=10,
               mutable=True,
               help=_('Sleep time between sending heartbeats.')),

    # Used for updating health
    cfg.StrOpt('health_update_driver', default='health_db',
               help=_('Driver for updating amphora health system.'),
               deprecated_for_removal=True,
               deprecated_reason=_('This driver interface was removed.'),
               deprecated_since='Victoria'),
]

oslo_messaging_opts = [
    cfg.StrOpt('topic'),
]

haproxy_amphora_opts = [
    cfg.StrOpt('base_path',
               default='/var/lib/octavia',
               help=_('Base directory for amphora files.')),
    cfg.StrOpt('base_cert_dir',
               default='/var/lib/octavia/certs',
               help=_('Base directory for cert storage.')),
    cfg.StrOpt('haproxy_template', help=_('Custom haproxy template.')),
    cfg.BoolOpt('connection_logging', default=True,
                help=_('Set this to False to disable connection logging.')),
    cfg.IntOpt('connection_max_retries',
               default=120,
               help=_('Retry threshold for connecting to amphorae.')),
    cfg.IntOpt('connection_retry_interval',
               default=5,
               help=_('Retry timeout between connection attempts in '
                      'seconds.')),
    cfg.IntOpt('active_connection_max_retries',
               default=15,
               help=_('Retry threshold for connecting to active amphorae.')),
    cfg.IntOpt('active_connection_rety_interval',
               default=2,
               help=_('Retry timeout between connection attempts in '
                      'seconds for active amphora.')),
    cfg.IntOpt('failover_connection_max_retries',
               default=2,
               help=_('Retry threshold for connecting to an amphora in '
                      'failover.')),
    cfg.IntOpt('failover_connection_retry_interval',
               default=5,
               help=_('Retry timeout between connection attempts in '
                      'seconds for amphora in failover.')),
    cfg.IntOpt('build_rate_limit',
               default=-1,
               help=_('Number of amphorae that could be built per controller '
                      'worker, simultaneously.')),
    cfg.IntOpt('build_active_retries',
               default=120,
               help=_('Retry threshold for waiting for a build slot for '
                      'an amphorae.')),
    cfg.IntOpt('build_retry_interval',
               default=5,
               help=_('Retry timeout between build attempts in '
                      'seconds.')),
    cfg.StrOpt('haproxy_stick_size', default='10k',
               help=_('Size of the HAProxy stick table. Accepts k, m, g '
                      'suffixes.  Example: 10k')),
    cfg.StrOpt('user_log_format',
               default='{{ project_id }} {{ lb_id }} %f %ci %cp %t %{+Q}r %ST '
                       '%B %U %[ssl_c_verify] %{+Q}[ssl_c_s_dn] %b %s %Tt '
                       '%tsc',
               help=_('Log format string for user flow logging.')),

    # REST server
    cfg.IPOpt('bind_host', default='::',  # nosec
              help=_("The host IP to bind to")),
    cfg.PortOpt('bind_port', default=9443,
                help=_("The port to bind to")),
    cfg.StrOpt('lb_network_interface',
               default='o-hm0',
               help=_('Network interface through which to reach amphora, only '
                      'required if using IPv6 link local addresses.')),
    cfg.StrOpt('haproxy_cmd', default='/usr/sbin/haproxy',
               help=_("The full path to haproxy")),
    cfg.IntOpt('respawn_count', default=2,
               help=_("The respawn count for haproxy's upstart script")),
    cfg.IntOpt('respawn_interval', default=2,
               help=_("The respawn interval for haproxy's upstart script")),
    cfg.FloatOpt('rest_request_conn_timeout', default=10,
                 help=_("The time in seconds to wait for a REST API "
                        "to connect.")),
    cfg.FloatOpt('rest_request_read_timeout', default=60,
                 help=_("The time in seconds to wait for a REST API "
                        "response.")),
    cfg.IntOpt('timeout_client_data',
               default=constants.DEFAULT_TIMEOUT_CLIENT_DATA,
               help=_('Frontend client inactivity timeout.')),
    cfg.IntOpt('timeout_member_connect',
               default=constants.DEFAULT_TIMEOUT_MEMBER_CONNECT,
               help=_('Backend member connection timeout.')),
    cfg.IntOpt('timeout_member_data',
               default=constants.DEFAULT_TIMEOUT_MEMBER_DATA,
               help=_('Backend member inactivity timeout.')),
    cfg.IntOpt('timeout_tcp_inspect',
               default=constants.DEFAULT_TIMEOUT_TCP_INSPECT,
               help=_('Time to wait for TCP packets for content inspection.')),
    # REST client
    cfg.StrOpt('client_cert', default='/etc/octavia/certs/client.pem',
               help=_("The client certificate to talk to the agent")),
    cfg.StrOpt('server_ca', default='/etc/octavia/certs/server_ca.pem',
               help=_("The ca which signed the server certificates")),
    cfg.BoolOpt('use_upstart', default=True,
                deprecated_for_removal=True,
                deprecated_reason='This is now automatically discovered '
                                  ' and configured.',
                help=_("If False, use sysvinit.")),
    cfg.IntOpt('api_db_commit_retry_attempts', default=15,
               help=_('The number of times the database action will be '
                      'attempted.')),
    cfg.IntOpt('api_db_commit_retry_initial_delay', default=1,
               help=_('The initial delay before a retry attempt.')),
    cfg.IntOpt('api_db_commit_retry_backoff', default=1,
               help=_('The time to backoff retry attempts.')),
    cfg.IntOpt('api_db_commit_retry_max', default=5,
               help=_('The maximum amount of time to wait between retry '
                      'attempts.')),
    cfg.IntOpt('default_connection_limit',
               default=constants.HAPROXY_DEFAULT_MAXCONN,
               help=_('Default connection_limit for listeners, used when '
                      'setting "-1" or when unsetting connection_limit with '
                      'the listener API.')),
]

controller_worker_opts = [
    cfg.IntOpt('workers',
               default=1, min=1,
               help='Number of workers for the controller-worker service.'),
    cfg.IntOpt('amp_active_retries',
               default=30,
               help=_('Retry attempts to wait for Amphora to become active')),
    cfg.IntOpt('amp_active_wait_sec',
               default=10,
               help=_('Seconds to wait between checks on whether an Amphora '
                      'has become active')),
    cfg.StrOpt('amp_flavor_id',
               default='',
               help=_('Nova instance flavor id for the Amphora')),
    cfg.StrOpt('amp_image_tag',
               default='',
               help=_('Glance image tag for the Amphora image to boot. '
                      'Use this option to be able to update the image '
                      'without reconfiguring Octavia.')),
    cfg.StrOpt('amp_image_owner_id',
               default='',
               help=_('Restrict glance image selection to a specific '
                      'owner ID.  This is a recommended security setting.')),
    cfg.StrOpt('amp_ssh_key_name',
               default='',
               help=_('Optional SSH keypair name, in nova, that will be used '
                      'for the authorized_keys inside the amphora.')),
    cfg.ListOpt('amp_boot_network_list',
                default='',
                help=_('List of networks to attach to the Amphorae. '
                       'All networks defined in the list will '
                       'be attached to each amphora.')),
    cfg.ListOpt('amp_secgroup_list',
                default='',
                help=_('List of security groups to attach to the Amphora.')),
    cfg.StrOpt('client_ca',
               default='/etc/octavia/certs/ca_01.pem',
               help=_('Client CA for the amphora agent to use')),
    cfg.StrOpt('amphora_driver',
               default='amphora_haproxy_rest_driver',
               help=_('Name of the amphora driver to use')),
    cfg.StrOpt('compute_driver',
               default='compute_nova_driver',
               help=_('Name of the compute driver to use')),
    cfg.StrOpt('network_driver',
               default='allowed_address_pairs_driver',
               help=_('Name of the network driver to use')),
    cfg.StrOpt('volume_driver',
               default=constants.VOLUME_NOOP_DRIVER,
               choices=constants.SUPPORTED_VOLUME_DRIVERS,
               help=_('Name of the volume driver to use')),
    cfg.StrOpt('image_driver',
               default='image_glance_driver',
               choices=constants.SUPPORTED_IMAGE_DRIVERS,
               help=_('Name of the image driver to use')),
    cfg.StrOpt('distributor_driver',
               default='distributor_noop_driver',
               help=_('Name of the distributor driver to use')),
    cfg.ListOpt('statistics_drivers', default=['stats_db'],
                deprecated_name='stats_update_driver',
                deprecated_group='health_manager',
                deprecated_since='Victoria',
                help=_('List of drivers for updating amphora statistics.')),
    cfg.StrOpt('loadbalancer_topology',
               default=constants.TOPOLOGY_SINGLE,
               choices=constants.SUPPORTED_LB_TOPOLOGIES,
               mutable=True,
               help=_('Load balancer topology configuration. '
                      'SINGLE - One amphora per load balancer. '
                      'ACTIVE_STANDBY - Two amphora per load balancer.')),
    cfg.BoolOpt('user_data_config_drive', default=False,
                help=_('If True, build cloud-init user-data that is passed '
                       'to the config drive on Amphora boot instead of '
                       'personality files. If False, utilize personality '
                       'files.')),
    cfg.IntOpt('amphora_delete_retries', default=5,
               help=_('Number of times an amphora delete should be retried.')),
    cfg.IntOpt('amphora_delete_retry_interval', default=5,
               help=_('Time, in seconds, between amphora delete retries.')),
]

task_flow_opts = [
    cfg.StrOpt('engine',
               default='parallel',
               choices=constants.SUPPORTED_TASKFLOW_ENGINE_TYPES,
               help=_('TaskFlow engine to use. '
                      'serial - Runs all tasks on a single thread. '
                      'parallel - Schedules tasks onto different threads to '
                      'allow for running non-dependent tasks simultaneously')),
    cfg.IntOpt('max_workers',
               default=5,
               help=_('The maximum number of workers')),
    cfg.BoolOpt('disable_revert', default=False,
                help=_('If True, disables the controller worker taskflow '
                       'flows from reverting.  This will leave resources in '
                       'an inconsistent state and should only be used for '
                       'debugging purposes.')),
    cfg.StrOpt('persistence_connection',
               default='sqlite://',
               help='Persistence database, which will be used to store tasks '
                    'states. Database connection url with db name'),
    cfg.BoolOpt('jobboard_enabled', default=False,
                help=_('If True, enables TaskFlow jobboard.')),
    cfg.StrOpt('jobboard_backend_driver',
               default='redis_taskflow_driver',
               choices=['redis_taskflow_driver', 'zookeeper_taskflow_driver'],
               help='Jobboard backend driver that will monitor job state.'),
    cfg.ListOpt('jobboard_backend_hosts', default=['127.0.0.1'],
                help='Jobboard backend server host(s).'),
    cfg.PortOpt('jobboard_backend_port', default=6379,
                help='Jobboard backend server port'),
    cfg.StrOpt('jobboard_backend_password', default='', secret=True,
               help='Jobboard backend server password'),
    cfg.StrOpt('jobboard_backend_namespace', default='octavia_jobboard',
               help='Jobboard name that should be used to store taskflow '
                    'job id and claims for it.'),
    cfg.StrOpt('jobboard_redis_sentinel', default=None,
               help='Sentinel name if it is used for Redis.'),
    cfg.DictOpt('jobboard_redis_backend_ssl_options',
                help='Redis jobboard backend ssl configuration options.',
                default={'ssl': False,
                         'ssl_keyfile': None,
                         'ssl_certfile': None,
                         'ssl_ca_certs': None,
                         'ssl_cert_reqs': 'required'}),
    cfg.DictOpt('jobboard_zookeeper_ssl_options',
                help='Zookeeper jobboard backend ssl configuration options.',
                default={'use_ssl': False,
                         'keyfile': None,
                         'keyfile_password': None,
                         'certfile': None,
                         'verify_certs': True}),
    cfg.IntOpt('jobboard_expiration_time', default=30,
               help='For backends like redis claiming jobs requiring setting '
                    'the expiry - how many seconds the claim should be '
                    'retained for.'),
    cfg.BoolOpt('jobboard_save_logbook', default=False,
                help='If for analysis required saving logbooks info, set this '
                     'parameter to True. By default remove logbook from '
                     'persistence backend when job completed.'),
]

core_cli_opts = []

certificate_opts = [
    cfg.StrOpt('cert_manager',
               default='barbican_cert_manager',
               help='Name of the cert manager to use'),
    cfg.StrOpt('cert_generator',
               default='local_cert_generator',
               help='Name of the cert generator to use'),
    cfg.StrOpt('barbican_auth',
               default='barbican_acl_auth',
               help='Name of the Barbican authentication method to use'),
    cfg.StrOpt('service_name',
               help=_('The name of the certificate service in the keystone '
                      'catalog')),
    cfg.StrOpt('endpoint', help=_('A new endpoint to override the endpoint '
                                  'in the keystone catalog.')),
    cfg.StrOpt('region_name',
               help='Region in Identity service catalog to use for '
                    'communication with the barbican service.'),
    cfg.StrOpt('endpoint_type',
               default='publicURL',
               help='The endpoint_type to be used for barbican service.'),
    cfg.StrOpt('ca_certificates_file',
               help=_('CA certificates file path')),
    cfg.BoolOpt('insecure',
                default=False,
                help=_('Disable certificate validation on SSL connections ')),
]

house_keeping_opts = [
    cfg.IntOpt('spare_check_interval',
               deprecated_for_removal=True,
               deprecated_since='Victoria',
               deprecated_reason='Spares Pool support will be removed in the '
                                 'X release.',
               default=30,
               help=_('Spare check interval in seconds')),
    cfg.IntOpt('spare_amphora_pool_size',
               deprecated_for_removal=True,
               deprecated_since='Victoria',
               deprecated_reason='Spares Pool support will be removed in the '
                                 'X release.',
               default=0,
               help=_('Number of spare amphorae')),
    cfg.IntOpt('cleanup_interval',
               default=30,
               help=_('DB cleanup interval in seconds')),
    cfg.IntOpt('amphora_expiry_age',
               default=604800,
               help=_('Amphora expiry age in seconds')),
    cfg.IntOpt('load_balancer_expiry_age',
               default=604800,
               help=_('Load balancer expiry age in seconds')),
    cfg.IntOpt('cert_interval',
               default=3600,
               help=_('Certificate check interval in seconds')),
    # 14 days for cert expiry buffer
    cfg.IntOpt('cert_expiry_buffer',
               default=1209600,
               help=_('Seconds until certificate expiration')),
    cfg.IntOpt('cert_rotate_threads',
               default=10,
               help=_('Number of threads performing amphora certificate'
                      ' rotation'))
]

keepalived_vrrp_opts = [
    cfg.IntOpt('vrrp_advert_int',
               default=1,
               help=_('Amphora role and priority advertisement interval '
                      'in seconds.')),
    cfg.IntOpt('vrrp_check_interval',
               default=5,
               help=_('VRRP health check script run interval in seconds.')),
    cfg.IntOpt('vrrp_fail_count',
               default=2,
               help=_('Number of successive failures before transition to a '
                      'fail state.')),
    cfg.IntOpt('vrrp_success_count',
               default=2,
               help=_('Number of consecutive successes before transition to a '
                      'success state.')),
    cfg.IntOpt('vrrp_garp_refresh_interval',
               default=5,
               help=_('Time in seconds between gratuitous ARP announcements '
                      'from the MASTER.')),
    cfg.IntOpt('vrrp_garp_refresh_count',
               default=2,
               help=_('Number of gratuitous ARP announcements to make on '
                      'each refresh interval.'))

]

nova_opts = [
    cfg.StrOpt('service_name',
               help=_('The name of the nova service in the keystone catalog')),
    cfg.StrOpt('endpoint', help=_('A new endpoint to override the endpoint '
                                  'in the keystone catalog.')),
    cfg.StrOpt('region_name',
               help=_('Region in Identity service catalog to use for '
                      'communication with the OpenStack services.')),
    cfg.StrOpt('endpoint_type', default='publicURL',
               help=_('Endpoint interface in identity service to use')),
    cfg.StrOpt('ca_certificates_file',
               help=_('CA certificates file path')),
    cfg.BoolOpt('insecure',
                default=False,
                help=_('Disable certificate validation on SSL connections')),
    cfg.BoolOpt('enable_anti_affinity', default=False,
                help=_('Flag to indicate if nova anti-affinity feature is '
                       'turned on. This option is only used when creating '
                       'amphorae in ACTIVE_STANDBY topology.')),
    cfg.StrOpt('anti_affinity_policy', default=constants.ANTI_AFFINITY,
               choices=[constants.ANTI_AFFINITY, constants.SOFT_ANTI_AFFINITY],
               help=_('Sets the anti-affinity policy for nova')),
    cfg.IntOpt('random_amphora_name_length', default=0,
               help=_('If non-zero, generate a random name of the length '
                      'provided for each amphora, in the format "a[A-Z0-9]*". '
                      'Otherwise, the default name format will be used: '
                      '"amphora-{UUID}".')),
    cfg.StrOpt('availability_zone', default=None,
               help=_('Availability zone to use for creating Amphorae')),
]

cinder_opts = [
    cfg.StrOpt('service_name',
               help=_('The name of the cinder service in the keystone '
                      'catalog')),
    cfg.StrOpt('endpoint', help=_('A new endpoint to override the endpoint '
                                  'in the keystone catalog.')),
    cfg.StrOpt('region_name',
               help=_('Region in Identity service catalog to use for '
                      'communication with the OpenStack services.')),
    cfg.StrOpt('endpoint_type', default='publicURL',
               help=_('Endpoint interface in identity service to use')),
    cfg.StrOpt('ca_certificates_file',
               help=_('CA certificates file path')),
    cfg.StrOpt('availability_zone', default=None,
               help=_('Availability zone to use for creating Volume')),
    cfg.BoolOpt('insecure',
                default=False,
                help=_('Disable certificate validation on SSL connections')),
    cfg.IntOpt('volume_size', default=16,
               help=_('Size of volume, in GB, for Amphora instance')),
    cfg.StrOpt('volume_type', default=None,
               help=_('Type of volume for Amphorae volume root disk')),
    cfg.IntOpt('volume_create_retry_interval', default=5,
               help=_('Interval time to wait volume is created in available '
                      'state')),
    cfg.IntOpt('volume_create_timeout', default=300,
               help=_('Timeout to wait for volume creation success')),
    cfg.IntOpt('volume_create_max_retries', default=5,
               help=_('Maximum number of retries to create volume'))
]

neutron_opts = [
    cfg.StrOpt('service_name',
               help=_('The name of the neutron service in the '
                      'keystone catalog')),
    cfg.StrOpt('endpoint', help=_('A new endpoint to override the endpoint '
                                  'in the keystone catalog.')),
    cfg.StrOpt('region_name',
               help=_('Region in Identity service catalog to use for '
                      'communication with the OpenStack services.')),
    cfg.StrOpt('endpoint_type', default='publicURL',
               help=_('Endpoint interface in identity service to use')),
    cfg.StrOpt('ca_certificates_file',
               help=_('CA certificates file path')),
    cfg.BoolOpt('insecure',
                default=False,
                help=_('Disable certificate validation on SSL connections ')),
]

glance_opts = [
    cfg.StrOpt('service_name',
               help=_('The name of the glance service in the '
                      'keystone catalog')),
    cfg.StrOpt('endpoint', help=_('A new endpoint to override the endpoint '
                                  'in the keystone catalog.')),
    cfg.StrOpt('region_name',
               help=_('Region in Identity service catalog to use for '
                      'communication with the OpenStack services.')),
    cfg.StrOpt('endpoint_type', default='publicURL',
               help=_('Endpoint interface in identity service to use')),
    cfg.StrOpt('ca_certificates_file',
               help=_('CA certificates file path')),
    cfg.BoolOpt('insecure',
                default=False,
                help=_('Disable certificate validation on SSL connections ')),
]

quota_opts = [
    cfg.IntOpt('default_load_balancer_quota',
               default=constants.QUOTA_UNLIMITED,
               help=_('Default per project load balancer quota.')),
    cfg.IntOpt('default_listener_quota',
               default=constants.QUOTA_UNLIMITED,
               help=_('Default per project listener quota.')),
    cfg.IntOpt('default_member_quota',
               default=constants.QUOTA_UNLIMITED,
               help=_('Default per project member quota.')),
    cfg.IntOpt('default_pool_quota',
               default=constants.QUOTA_UNLIMITED,
               help=_('Default per project pool quota.')),
    cfg.IntOpt('default_health_monitor_quota',
               default=constants.QUOTA_UNLIMITED,
               help=_('Default per project health monitor quota.')),
    cfg.IntOpt('default_l7policy_quota',
               default=constants.QUOTA_UNLIMITED,
               help=_('Default per project l7policy quota.')),
    cfg.IntOpt('default_l7rule_quota',
               default=constants.QUOTA_UNLIMITED,
               help=_('Default per project l7rule quota.')),
]

audit_opts = [
    cfg.BoolOpt('enabled', default=False,
                help=_('Enable auditing of API requests')),
    cfg.StrOpt('audit_map_file',
               default='/etc/octavia/octavia_api_audit_map.conf',
               help=_('Path to audit map file for octavia-api service. '
                      'Used only when API audit is enabled.')),
    cfg.StrOpt('ignore_req_list', default='',
               help=_('Comma separated list of REST API HTTP methods to be '
                      'ignored during audit. For example: auditing will not '
                      'be done on any GET or POST requests if this is set to '
                      '"GET,POST". It is used only when API audit is '
                      'enabled.')),
]

driver_agent_opts = [
    cfg.StrOpt('status_socket_path',
               default='/var/run/octavia/status.sock',
               help=_('Path to the driver status unix domain socket file.')),
    cfg.StrOpt('stats_socket_path',
               default='/var/run/octavia/stats.sock',
               help=_('Path to the driver statistics unix domain socket '
                      'file.')),
    cfg.StrOpt('get_socket_path',
               default='/var/run/octavia/get.sock',
               help=_('Path to the driver get unix domain socket file.')),
    cfg.IntOpt('status_request_timeout',
               default=5,
               help=_('Time, in seconds, to wait for a status update '
                      'request.')),
    cfg.IntOpt('status_max_processes',
               default=50,
               help=_('Maximum number of concurrent processes to use '
                      'servicing status updates.')),
    cfg.IntOpt('stats_request_timeout',
               default=5,
               help=_('Time, in seconds, to wait for a statistics update '
                      'request.')),
    cfg.IntOpt('stats_max_processes',
               default=50,
               help=_('Maximum number of concurrent processes to use '
                      'servicing statistics updates.')),
    cfg.IntOpt('get_request_timeout',
               default=5,
               help=_('Time, in seconds, to wait for a get request.')),
    cfg.IntOpt('get_max_processes',
               default=50,
               help=_('Maximum number of concurrent processes to use '
                      'servicing get requests.')),
    cfg.FloatOpt('max_process_warning_percent',
                 default=0.75, min=0.01, max=0.99,
                 help=_('Percentage of max_processes (both status and stats) '
                        'in use to start logging warning messages about an '
                        'overloaded driver-agent.')),
    cfg.IntOpt('provider_agent_shutdown_timeout',
               default=60,
               help=_('The time, in seconds, to wait for provider agents '
                      'to shutdown after the exit event has been set.')),
    cfg.ListOpt('enabled_provider_agents', default='',
                help=_('List of enabled provider agents. The driver-agent '
                       'will launch these agents at startup.'))
]

# Register the configuration options
cfg.CONF.register_opts(core_opts)
cfg.CONF.register_opts(api_opts, group='api_settings')
cfg.CONF.register_opts(amphora_agent_opts, group='amphora_agent')
cfg.CONF.register_opts(compute_opts, group='compute')
cfg.CONF.register_opts(networking_opts, group='networking')
cfg.CONF.register_opts(oslo_messaging_opts, group='oslo_messaging')
cfg.CONF.register_opts(haproxy_amphora_opts, group='haproxy_amphora')
cfg.CONF.register_opts(controller_worker_opts, group='controller_worker')
cfg.CONF.register_opts(keepalived_vrrp_opts, group='keepalived_vrrp')
cfg.CONF.register_opts(task_flow_opts, group='task_flow')
cfg.CONF.register_opts(house_keeping_opts, group='house_keeping')
cfg.CONF.register_opts(certificate_opts, group='certificates')
cfg.CONF.register_opts(health_manager_opts, group='health_manager')
cfg.CONF.register_opts(nova_opts, group='nova')
cfg.CONF.register_opts(cinder_opts, group='cinder')
cfg.CONF.register_opts(glance_opts, group='glance')
cfg.CONF.register_opts(neutron_opts, group='neutron')
cfg.CONF.register_opts(quota_opts, group='quotas')
cfg.CONF.register_opts(audit_opts, group='audit')
cfg.CONF.register_opts(driver_agent_opts, group='driver_agent')

cfg.CONF.register_opts(local.certgen_opts, group='certificates')
cfg.CONF.register_opts(local.certmgr_opts, group='certificates')

# Ensure that the control exchange is set correctly
messaging.set_transport_defaults(control_exchange='octavia')
_SQL_CONNECTION_DEFAULT = 'sqlite://'
# Update the default QueuePool parameters. These can be tweaked by the
# configuration variables - max_pool_size, max_overflow and pool_timeout
db_options.set_defaults(cfg.CONF, connection=_SQL_CONNECTION_DEFAULT,
                        max_pool_size=10, max_overflow=20, pool_timeout=10)

ks_loading.register_auth_conf_options(cfg.CONF, constants.SERVICE_AUTH)
ks_loading.register_session_conf_options(cfg.CONF, constants.SERVICE_AUTH)


def register_cli_opts():
    cfg.CONF.register_cli_opts(core_cli_opts)
    logging.register_options(cfg.CONF)


def init(args, **kwargs):
    register_cli_opts()
    cfg.CONF(args=args, project='octavia',
             version='%%prog %s' % version.version_info.release_string(),
             **kwargs)
    validate.check_default_tls_versions_min_conflict()
    setup_remote_debugger()
    validate.check_default_ciphers_prohibit_list_conflict()


def setup_logging(conf):
    """Sets up the logging options for a log with supplied name.

    :param conf: a cfg.ConfOpts object
    """
    logging.set_defaults(default_log_levels=logging.get_default_log_levels() +
                         EXTRA_LOG_LEVEL_DEFAULTS)
    product_name = "octavia"
    logging.setup(conf, product_name)
    LOG.info("Logging enabled!")
    LOG.info("%(prog)s version %(version)s",
             {'prog': sys.argv[0],
              'version': version.version_info.release_string()})
    LOG.debug("command line: %s", " ".join(sys.argv))


def _enable_pydev(debugger_host, debugger_port):
    try:
        from pydev import pydevd  # pylint: disable=import-outside-toplevel
    except ImportError:
        import pydevd  # pylint: disable=import-outside-toplevel

    pydevd.settrace(debugger_host,
                    port=int(debugger_port),
                    stdoutToServer=True,
                    stderrToServer=True)


def _enable_ptvsd(debuggger_host, debugger_port):
    import ptvsd  # pylint: disable=import-outside-toplevel

    # Allow other computers to attach to ptvsd at this IP address and port.
    ptvsd.enable_attach(address=(debuggger_host, debugger_port),
                        redirect_output=True)

    # Pause the program until a remote debugger is attached
    ptvsd.wait_for_attach()


def setup_remote_debugger():
    """Required setup for remote debugging."""

    debugger_type = os.environ.get('DEBUGGER_TYPE', 'pydev')
    debugger_host = os.environ.get('DEBUGGER_HOST')
    debugger_port = os.environ.get('DEBUGGER_PORT')

    if not debugger_type or not debugger_host or not debugger_port:
        return

    try:
        LOG.warning("Connecting to remote debugger. Once connected, resume "
                    "the program on the debugger to continue with the "
                    "initialization of the service.")
        if debugger_type == 'pydev':
            _enable_pydev(debugger_host, debugger_port)
        elif debugger_type == 'ptvsd':
            _enable_ptvsd(debugger_host, debugger_port)
        else:
            LOG.exception('Debugger %(debugger)s is not supported',
                          debugger_type)
    except Exception:
        LOG.exception('Unable to join debugger, please make sure that the '
                      'debugger processes is listening on debug-host '
                      '\'%(debug-host)s\' debug-port \'%(debug-port)s\'.',
                      {'debug-host': debugger_host,
                       'debug-port': debugger_port})
        raise


def set_lib_defaults():
    """Update default value for configuration options from other namespace.

    Example, oslo lib config options. This is needed for
    config generator tool to pick these default value changes.
    https://docs.openstack.org/oslo.config/latest/cli/
    generator.html#modifying-defaults-from-other-namespaces
    """

    # TODO(gmann): Remove setting the default value of config policy_file
    # once oslo_policy change the default value to 'policy.yaml'.
    # https://github.com/openstack/oslo.policy/blob/a626ad12fe5a3abd49d70e3e5b95589d279ab578/oslo_policy/opts.py#L49
    # Update default value of oslo.policy policy_file config option.
    policy_opts.set_defaults(cfg.CONF, 'policy.yaml')
